import os

import cv2
import numpy as np
from PySide6 import QtCore, QtWidgets

from piscat.GUI.InputOutput import import_im2vid, import_RAW, video_cropping
from piscat.GUI.Visualization.fun_display_localization import Visulization_localization
from piscat.InputOutput import image_to_video, read_status_line, reading_videos


class Reading(QtWidgets.QMainWindow):
    update_output = QtCore.Signal(object)

    def __init__(self):
        super(Reading, self).__init__()
        self.original_video = None
        self.filename = None
        self.folder = None

    def askdir_file(self):
        self.filename = False
        dialog = QtWidgets.QFileDialog(self)
        dialog.setWindowTitle("Open File")
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            self.filename = dialog.selectedFiles()[0]
        if type(self.filename) is str:
            file_extention = os.path.splitext(self.filename)[1]
            if (
                file_extention == ".pf"
                or file_extention == ".raw"
                or file_extention == ".bin"
                or file_extention == ".PF"
                or file_extention == ".RAW"
                or file_extention == ".BIN"
            ):
                return "Raw"
            elif (
                file_extention == ".jpg"
                or file_extention == ".JPG"
                or file_extention == ".png"
                or file_extention == ".PNG"
            ):
                return "PNG"
            elif file_extention == ".tiff" or file_extention == ".TIFF":
                return "TIF"
            elif file_extention == ".avi" or file_extention == ".AVI":
                return "AVI"
            elif file_extention == ".TIF" or file_extention == ".tif":
                return "TIF"
            elif file_extention == ".fits" or file_extention == ".fits":
                return "Fits"
            elif file_extention == ".fli" or file_extention == ".fli":
                return "Fli"
            elif file_extention == ".py" or file_extention == ".PY":
                return "Python"

    def askdir(self):
        folder = QtWidgets.QFileDialog.getExistingDirectory(
            None, "Select a folder:", "", QtWidgets.QFileDialog.ShowDirsOnly
        )
        return folder

    def read_video(self):
        title = self.askdir_file()
        if self.filename:
            if title == "Raw":
                self.image = import_RAW.RawImage(self.filename)
                while self.image.raw_data_update_flag:
                    QtWidgets.QApplication.processEvents()
                try:
                    video = reading_videos.read_binary(
                        self.filename,
                        img_width=self.image.width_size,
                        img_height=self.image.height_size,
                        image_type=self.image.set_bit_order + self.image.type,
                    )

                    if self.image.groupBox_cropping.isChecked():
                        self.original_video = video[
                            self.image.frame_s : self.image.frame_e : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]

                    else:
                        self.original_video = video

                    self.original_video, status_line_info = self.status_line_remove(
                        self.original_video
                    )

                    if self.image.flag_display is True:
                        self.visualization_ = Visulization_localization()
                        self.visualization_.new_display(
                            self.original_video,
                            self.original_video,
                            object=None,
                            title="Raw video",
                        )

                    self.update_output.emit(
                        [self.original_video, title, self.filename, status_line_info]
                    )
                except:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Selected parameters are not correct!")
                    self.msg_box.exec_()

            elif title == "PNG":
                png_img = reading_videos.read_png(self.filename)
                self.original_video = png_img

                self.visualization_ = Visulization_localization(self.filename)
                self.visualization_.new_display(
                    self.original_video, self.original_video, object=None, title="PNG"
                )

                self.update_output.emit([self.original_video, title, self.filename, None])

            elif title == "AVI":
                avi_video = reading_videos.read_avi(self.filename)

                self.image = video_cropping.Cropping(self)

                while self.image.raw_data_update_flag:
                    QtWidgets.QApplication.processEvents()

                if self.image.frame_e is not None:
                    if self.image.frame_e != -1:
                        self.original_video = avi_video[
                            self.image.frame_s : self.image.frame_e : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                    elif self.image.frame_e == -1:
                        self.original_video = avi_video[
                            self.image.frame_s : -1 : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                else:
                    self.original_video = avi_video

                self.original_video = self.original_video.copy(order="C")

                if self.image.flag_display is True:
                    self.visualization_ = Visulization_localization()
                    self.visualization_.new_display(
                        self.original_video, self.original_video, object=None, title="AVI"
                    )

                self.update_output.emit([self.original_video, title, self.filename, None])

            elif title == "TIF":
                tif_video = reading_videos.read_tif(self.filename)
                self.image = video_cropping.Cropping(self)

                while self.image.raw_data_update_flag:
                    QtWidgets.QApplication.processEvents()

                if self.image.flag_RGB2GRAY:
                    if tif_video.shape[2] == 4:
                        tif_video = cv2.cvtColor(tif_video, cv2.COLOR_BGR2GRAY)
                        tif_video = np.expand_dims(tif_video, axis=0)

                if self.image.frame_e is not None:
                    if self.image.frame_e != -1:
                        self.original_video = tif_video[
                            self.image.frame_s : self.image.frame_e : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                    elif self.image.frame_e == -1:
                        self.original_video = tif_video[
                            self.image.frame_s :: self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                else:
                    self.original_video = tif_video

                self.original_video = self.original_video.copy(order="C")

                if self.image.flag_display is True:
                    self.visualization_ = Visulization_localization()
                    self.visualization_.new_display(
                        self.original_video, self.original_video, object=None, title="TIF"
                    )

                self.update_output.emit([self.original_video, title, self.filename, None])

            elif title == "Fits":
                fits_video = reading_videos.read_fits(self.filename)
                self.image = video_cropping.Cropping(self)

                while self.image.raw_data_update_flag:
                    QtWidgets.QApplication.processEvents()

                if self.image.frame_e is not None:
                    if self.image.frame_e != -1:
                        self.original_video = fits_video[
                            self.image.frame_s : self.image.frame_e : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                    elif self.image.frame_e == -1:
                        self.original_video = fits_video[
                            self.image.frame_s :: self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                else:
                    self.original_video = fits_video

                self.original_video = self.original_video.copy(order="C")

                if self.image.flag_display is True:
                    self.visualization_ = Visulization_localization()
                    self.visualization_.new_display(
                        self.original_video, self.original_video, object=None, title="TIF"
                    )

                self.update_output.emit([self.original_video, title, self.filename])

            elif title == "Fli":
                fli_video = reading_videos.read_fli(self.filename)
                self.image = video_cropping.Cropping(self)

                while self.image.raw_data_update_flag:
                    QtWidgets.QApplication.processEvents()

                if self.image.frame_e is not None:
                    if self.image.frame_e != -1:
                        self.original_video = fli_video[
                            self.image.frame_s : self.image.frame_e : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                    elif self.image.frame_e == -1:
                        self.original_video = fli_video[
                            self.image.frame_s :: self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                else:
                    self.original_video = fli_video

                self.original_video = self.original_video.copy(order="C")

                if self.image.flag_display is True:
                    self.visualization_ = Visulization_localization()
                    self.visualization_.new_display(
                        self.original_video, self.original_video, object=None, title="TIF"
                    )

                self.update_output.emit([self.original_video, title, self.filename])

            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("The video type is not defined!")
                self.msg_box.exec_()

        elif self.filename is False:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Filename is not valid.!")
            self.msg_box.exec_()

    def load_batch_data(self):
        data_setting = {}
        self.filename = self.askdir()
        self.image = import_RAW.RawImage(self.filename)
        while self.image.raw_data_update_flag:
            QtWidgets.QApplication.processEvents()
        try:
            data_setting["path"] = self.filename
            data_setting["title"] = "Raw"
            data_setting["img_width"] = self.image.width_size
            data_setting["img_height"] = self.image.height_size
            data_setting["image_type"] = self.image.set_bit_order + self.image.type
            data_setting["s_frame"] = self.image.frame_s
            data_setting["e_frame"] = self.image.frame_e

            if self.image.groupBox_cropping.isChecked():
                data_setting["frame_stride"] = self.image.frame_jump
                data_setting["width_size_s"] = self.image.width_size_s
                data_setting["width_size_e"] = self.image.width_size_e
                data_setting["height_size_s"] = self.image.height_size_s
                data_setting["height_size_e"] = self.image.height_size_e
            else:
                data_setting["frame_stride"] = 1
                data_setting["width_size_s"] = 0
                data_setting["width_size_e"] = -1
                data_setting["height_size_s"] = 0
                data_setting["height_size_e"] = -1
            self.update_output.emit([data_setting])
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Selected parameters are not correct!")
            self.msg_box.exec_()

    def im2video(self):
        title = "image2video"
        folder = self.askdir()
        self.image = import_im2vid.Image2Video(self)
        while self.image.raw_data_update_flag:
            QtWidgets.QApplication.processEvents()
        try:
            if self.image.im_type is not None:
                im2vid = image_to_video.Image2Video(
                    path=folder,
                    file_format=self.image.im_type,
                    width_size=self.image.width_size,
                    height_size=self.image.height_size,
                    image_type=np.dtype(self.image.set_bit_order + self.image.type),
                    reader_type=self.image.video_reader_type,
                )
                video = im2vid()

                video, status_line_info = self.status_line_remove(video)

                self.original_video = video

                if self.image.flag_display is True:
                    self.visualization_ = Visulization_localization()
                    self.visualization_.new_display(
                        self.original_video, self.original_video, object=None, title="im2video"
                    )

                self.update_output.emit(
                    [self.original_video, title, self.folder, status_line_info]
                )
            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Type of image is not defined!")
                self.msg_box.exec_()
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Selected parameters are not correct!")
            self.msg_box.exec_()

    def tiff_video(self):
        title = self.askdir_file()

        if self.filename:
            if title == "TIF":
                tif_video = reading_videos.read_tif_iterate(self.filename)
                self.image = video_cropping.Cropping(self)

                while self.image.raw_data_update_flag:
                    QtWidgets.QApplication.processEvents()

                if self.image.flag_RGB2GRAY:
                    if tif_video.shape[2] == 4:
                        tif_video = cv2.cvtColor(tif_video, cv2.COLOR_BGR2GRAY)
                        tif_video = np.expand_dims(tif_video, axis=0)

                if self.image.frame_e is not None:
                    if self.image.frame_e != -1:
                        self.original_video = tif_video[
                            self.image.frame_s : self.image.frame_e : self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                    elif self.image.frame_e == -1:
                        self.original_video = tif_video[
                            self.image.frame_s :: self.image.frame_jump,
                            self.image.width_size_s : self.image.width_size_e,
                            self.image.height_size_s : self.image.height_size_e,
                        ]
                else:
                    self.original_video = tif_video

                self.original_video = self.original_video.copy(order="C")

                if self.image.flag_display is True:
                    self.visualization_ = Visulization_localization()
                    self.visualization_.new_display(
                        self.original_video, self.original_video, object=None, title="TIF"
                    )

                self.update_output.emit([self.original_video, title, self.filename, None])

            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Type of image is not defined!")
                self.msg_box.exec_()

    def run_py_script(self):
        title = self.askdir_file()

        if title == "Python":
            self.threadpool.start(self.run_py_script_thread(self.filename))
            self.set_plain_text(" Done")

        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("File bin_type is not valid.!")
            self.msg_box.exec_()

    def status_line_remove(self, video):
        statusLine = read_status_line.StatusLine(video)
        cut_frames, axis_status_line = statusLine.find_status_line()

        if axis_status_line == "row" or axis_status_line == "column":
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText(
                "status line remove"
                + "\nold shape:"
                + str(video.shape)
                + "\nnew shape:"
                + str(cut_frames.shape)
            )
            self.msg_box.exec_()

        return cut_frames, axis_status_line

    def closeEvent(self, event):
        QtCore.QCoreApplication.instance().quit()
        print("closing PlaySetting")



import cv2 as cv
from skimage import exposure


class ContrastAdjustment:
    def __init__(self, video):
        """This class is used in the GUI to change the visualization's
        brightness and contrast.

        Parameters
        ----------
        video: NDArray
            Input video.
        """
        self.video = video

    def pixel_transforms(self, image, alpha, beta, min_intensity, max_intensity):
        """Using the value of the hyperparameters, adjust the brightness and
        contrast of the current image.

        Parameters
        ----------
        image: NDArray
            Input image (2D-Numpy).

        alpha: float
            Scale factor.

        beta: float
            Delta added to the scaled values.

        min_intensity: float
            Min intensity values of output image.

        max_intensity: float
            Max intensity values of output image.

        Returns
        -------
        image_: NDArray
            Output image (2D-Numpy)
        """
        video_rescale = exposure.rescale_intensity(
            image, out_range=(min_intensity, max_intensity)
        )
        image_ = cv.convertScaleAbs(video_rescale, alpha=alpha, beta=beta)
        return image_

    def auto_pixelTransforms(self, image):
        """
        Adjusting the brightness and contrast of the current image automatically.

        Parameters
        ----------
        image: NDArray
            Input image (2D-Numpy).

        Returns
        -------
        image_: NDArray
            Output image (2D-Numpy)
        """
        min_intensity = image.min()
        max_intensity = image.max()
        alpha = 255 / (max_intensity - min_intensity)
        beta = -min_intensity * alpha
        return self.pixel_transforms(image, alpha, beta, min_intensity, max_intensity)



import numpy as np


def fixed_length(list_of_lists):
    """
    This function takes a list of lists of length N, pads all but the longest
    sublists with trailing zeros so that they have the same length K, and returns
    an array with shape (N, K) containing all the data of these lists.

    Parameters
    ----------
    list_of_lists: list[list]
        A list of lists.

    Returns
    -------
    array: np.ndarray
    """
    array = np.zeros((len(list_of_lists), max(len(l) for l in list_of_lists)))
    for index, row in enumerate(list_of_lists):
        array[index][0 : len(row)] = row
    return array


def protein_trajectories_list2dic(v_shape_list):
    """From the output of ``TemporalFilter.v_trajectory``, this function
    converts the list to dictionary format.

    Parameters
    ----------
    v_shape_list: List of list

        | [intensity_horizontal, intensity_vertical, particle_center_intensity,
           particle_center_intensity_follow, particle_frame, particle_sigma,
           particle_X, particle_Y, particle_ID, optional(fit_intensity, fit_x,
           fit_y, fit_X_sigma, fit_Y_sigma, fit_Bias, fit_intensity_error,
           fit_x_error, fit_y_error, fit_X_sigma_error, fit_Y_sigma_error,
           fit_Bias_error)]

    Returns
    -------
    dic_all: dic
       Return dictionary similar to the following structures

       | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ..., ...,
          'particle_ID': ...}, "#1": {}, ...}

    """
    dic_all = {}

    if type(v_shape_list) is list:
        v_shape_list = np.asarray(v_shape_list)

    if type(v_shape_list) is np.ndarray:
        for n_particle in range(v_shape_list.shape[0]):
            dic_particles = {}

            if type(v_shape_list[n_particle][0]) is list:
                dic_particles["intensity_horizontal"] = np.asarray(
                    fixed_length(v_shape_list[n_particle][0])
                )
                dic_particles["intensity_vertical"] = np.asarray(
                    fixed_length(v_shape_list[n_particle][1])
                )
                dic_particles["center_int"] = np.asarray(v_shape_list[n_particle][2])
                dic_particles["center_int_flow"] = np.asarray(v_shape_list[n_particle][3])
                dic_particles["frame_number"] = np.asarray(v_shape_list[n_particle][4])
                dic_particles["sigma"] = np.asarray(v_shape_list[n_particle][5])
                dic_particles["x_center"] = np.asarray(v_shape_list[n_particle][6])
                dic_particles["y_center"] = np.asarray(v_shape_list[n_particle][7])
                dic_particles["particle_ID"] = np.asarray(v_shape_list[n_particle][8])
            else:
                dic_particles["intensity_horizontal"] = v_shape_list[n_particle][0].ravel()
                dic_particles["intensity_vertical"] = v_shape_list[n_particle][1].ravel()
                dic_particles["center_int"] = v_shape_list[n_particle][2].ravel()
                dic_particles["center_int_flow"] = v_shape_list[n_particle][3].ravel()
                dic_particles["frame_number"] = v_shape_list[n_particle][4].ravel()
                dic_particles["sigma"] = v_shape_list[n_particle][5].ravel()
                dic_particles["x_center"] = v_shape_list[n_particle][6].ravel()
                dic_particles["y_center"] = v_shape_list[n_particle][7].ravel()
                dic_particles["particle_ID"] = v_shape_list[n_particle][8].ravel()

            num_parameters = len(v_shape_list[n_particle])

            if num_parameters == 21:
                if type(v_shape_list[n_particle][0]) is list:
                    dic_particles["fit_intensity"] = np.asarray(v_shape_list[n_particle][9])
                    dic_particles["fit_x"] = np.asarray(v_shape_list[n_particle][10])
                    dic_particles["fit_y"] = np.asarray(v_shape_list[n_particle][11])
                    dic_particles["fit_X_sigma"] = np.asarray(v_shape_list[n_particle][12])
                    dic_particles["fit_Y_sigma"] = np.asarray(v_shape_list[n_particle][13])
                    dic_particles["fit_Bias"] = np.asarray(v_shape_list[n_particle][14])
                    dic_particles["fit_intensity_error"] = np.asarray(
                        v_shape_list[n_particle][15]
                    )
                    dic_particles["fit_x_error"] = np.asarray(v_shape_list[n_particle][16])
                    dic_particles["fit_y_error"] = np.asarray(v_shape_list[n_particle][17])
                    dic_particles["fit_X_sigma_error"] = np.asarray(v_shape_list[n_particle][18])
                    dic_particles["fit_Y_sigma_error"] = np.asarray(v_shape_list[n_particle][19])
                    dic_particles["fit_Bias_error"] = np.asarray(v_shape_list[n_particle][20])

                else:
                    dic_particles["fit_intensity"] = v_shape_list[n_particle][9].ravel()
                    dic_particles["fit_x"] = v_shape_list[n_particle][10].ravel()
                    dic_particles["fit_y"] = v_shape_list[n_particle][11].ravel()
                    dic_particles["fit_X_sigma"] = v_shape_list[n_particle][12].ravel()
                    dic_particles["fit_Y_sigma"] = v_shape_list[n_particle][13].ravel()
                    dic_particles["fit_Bias"] = v_shape_list[n_particle][14].ravel()
                    dic_particles["fit_intensity_error"] = v_shape_list[n_particle][15].ravel()
                    dic_particles["fit_x_error"] = v_shape_list[n_particle][16].ravel()
                    dic_particles["fit_y_error"] = v_shape_list[n_particle][17].ravel()
                    dic_particles["fit_X_sigma_error"] = v_shape_list[n_particle][18].ravel()
                    dic_particles["fit_Y_sigma_error"] = v_shape_list[n_particle][19].ravel()
                    dic_particles["fit_Bias_error"] = v_shape_list[n_particle][20].ravel()
            dic_all["#" + str(n_particle)] = dic_particles
        return dic_all



import datetime
import multiprocessing
import os
import subprocess
import sys
import time
import warnings
from functools import partial

import pkg_resources
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QThreadPool

from piscat import version
from piscat.GUI.BatchAnalysis.batch_data import BatchAnalysis
from piscat.GUI.CPU_Configurations import CPU_setting_wrapper
from piscat.GUI.InputOutput import Reading
from piscat.GUI.InputOutput.help import Help
from piscat.GUI.Memory.video_in_memory import VideoInMemory
from piscat.GUI.Proccessing import FFT2D_GUI_wrapper
from piscat.GUI.ProgressBar.fun_progressBar import ProgressBar
from piscat.GUI.Projects.iPSF.iPSF_model import GUI_iPSF
from piscat.GUI.Projects.Protein import Noise_Floor, ProteinTabs
from piscat.GUI.VideoAnalysis import Analysis, AnalysisConstant
from piscat.GUI.Visualization.fun_display_localization import Visulization_localization

warnings.filterwarnings("ignore")


class PiSCAT_GUI(QtWidgets.QMainWindow):
    progressChanged = QtCore.Signal(int)

    def __init__(self):
        super(PiSCAT_GUI, self).__init__()

        self.threadpool = QThreadPool()
        self.original_video = None
        self.dra_video = None

        self.FFT2D_GUI_wrapper = FFT2D_GUI_wrapper()

        self.initUI()

    def initUI(self):
        # defining the main window.
        window_icon = pkg_resources.resource_filename("piscat.GUI.icons", "mpl.png")
        main_icon = QtGui.QIcon(window_icon)
        self.setWindowIcon(main_icon)
        self.setWindowTitle("PiSCAT")
        self.setStyleSheet("QMainWindow{background-color: darkgray;}")
        self.setGeometry(90, 90, 400, 1000)
        self.setFixedSize(340, 600)

        # defining the main frame as the central widget and give it a vertical layout.
        self.main_frame = QtWidgets.QFrame()
        self.setCentralWidget(self.main_frame)
        self.main_frame_layout = QtWidgets.QVBoxLayout()
        self.main_frame.setLayout(self.main_frame_layout)

        # ---------- Adding the main Logo ----------
        label_logo = QtWidgets.QLabel(self)
        logo_piscat_path = pkg_resources.resource_filename(
            "piscat.GUI.icons", "PiSCAT_logo_bg.png"
        )
        loader_logo = QtGui.QPixmap(logo_piscat_path)
        loader_logo = loader_logo.scaled(320, 100)
        label_logo.setPixmap(loader_logo)
        self.main_frame_layout.addWidget(label_logo)

        label = QtWidgets.QLabel(self)
        movie_path = pkg_resources.resource_filename("piscat.GUI.icons", "movie.gif")
        loader_gif = QtGui.QMovie(movie_path)
        self.resize(loader_gif.currentPixmap().width(), loader_gif.currentPixmap().height())
        loader_gif.start()
        label.setMovie(loader_gif)
        self.main_frame_layout.addWidget(label)

        # ---------- defining the toolbox GroupBox and add it to the main frame. ----------
        self.toolbox_GB = QtWidgets.QGroupBox("Information")
        self.toolbox_GB.setFixedHeight(200)
        self.toolbox_GB_layout = QtWidgets.QVBoxLayout()
        self.toolbox_GB.setLayout(self.toolbox_GB_layout)
        self.main_frame_layout.addWidget(self.toolbox_GB)

        self.textEdit = QtWidgets.QTextEdit()
        self.textEdit.setReadOnly(True)
        self.textEdit.verticalScrollBar().minimum()
        # textEdit.setTextBackgroundColor(QtGui.QColor(140,140,0))
        self.toolbox_GB_layout.addWidget(self.textEdit)

        # adding different tab of "File" in menuBar.
        menubar = self.menuBar()

        # --------File------
        fileMenu = menubar.addMenu("&File")

        open_file = QtGui.QAction(QtGui.QIcon("folder.ico"), "Open", self)
        # open_file.setShortcut('Ctrl+S')
        open_file.setStatusTip("Open new File")
        fileMenu.addAction(open_file)
        self.connect(open_file, QtCore.SIGNAL("triggered()"), self.video_loading_wrapper)

        imp = fileMenu.addMenu("Import")
        open_im2vid = imp.addAction("Image Sequence to video")
        self.connect(open_im2vid, QtCore.SIGNAL("triggered()"), self.image2video_loading_wrapper)

        open_tiff_vid = imp.addAction("TIFF")
        self.connect(open_tiff_vid, QtCore.SIGNAL("triggered()"), self.tiff_video_loading_wrapper)

        load_python_script = imp.addAction("Run Python script")
        self.connect(load_python_script, QtCore.SIGNAL("triggered()"), Reading().run_py_script)

        Batch_analysis = imp.addAction("Run Batch analysis")
        self.connect(Batch_analysis, QtCore.SIGNAL("triggered()"), self.batch_analysis)
        # --------File------

        # --------Setting--------
        setting_menu = menubar.addMenu("&Setting")

        setting_cpu = QtGui.QAction(QtGui.QIcon("folder.ico"), "CPU_Configurations", self)
        # setting_cpu.setShortcut('Ctrl+O')
        setting_cpu.setStatusTip("Open new File")
        setting_menu.addAction(setting_cpu)
        self.connect(setting_cpu, QtCore.SIGNAL("triggered()"), self.cpu_setting)
        # --------Setting--------

        # --------View --------
        self.list_available_video = {}
        view_menu = menubar.addMenu("&View")

        self.video_in_memory_flag = {
            "original_video": False,
            "DRA_video": False,
            "Protein_Tracking": False,
            "GUV3D": False,
            "arc_cos": False,
            "PSF Clustering": False,
            "Template_matching": False,
        }

        Display_video_in_memory = QtGui.QAction(QtGui.QIcon("folder.ico"), "Loaded videos", self)
        # open_file.setShortcut('Ctrl+S')
        Display_video_in_memory.setStatusTip("Display")
        view_menu.addAction(Display_video_in_memory)
        self.connect(
            Display_video_in_memory, QtCore.SIGNAL("triggered()"), self.available_video_in_memory
        )
        # --------View --------

        # --------Process--------
        process_menu = menubar.addMenu("&Process")

        spectrum = QtGui.QAction(QtGui.QIcon("folder.ico"), "FFT", self)
        # open_file.setShortcut('Ctrl+S')
        spectrum.setStatusTip("FFT")
        process_menu.addAction(spectrum)
        self.connect(spectrum, QtCore.SIGNAL("triggered()"), self.spectrum_init)

        image_calculator = QtGui.QAction(QtGui.QIcon("folder.ico"), "Image calculator", self)
        # open_file.setShortcut('Ctrl+S')
        image_calculator.setStatusTip("Image calculator")
        process_menu.addAction(image_calculator)
        self.connect(image_calculator, QtCore.SIGNAL("triggered()"), self.analysis_wrapper)

        image_calculator_2 = QtGui.QAction(
            QtGui.QIcon("folder.ico"), "Image calculator constant number", self
        )
        # open_file.setShortcut('Ctrl+S')
        image_calculator_2.setStatusTip("Image calculator constant number")
        process_menu.addAction(image_calculator_2)
        self.connect(image_calculator_2, QtCore.SIGNAL("triggered()"), self.analysis_wrapper_2)
        # --------Process--------

        # --------Analyze.--------
        analyze_menu = menubar.addMenu("&Analyze")

        noise_floor = QtGui.QAction("Noise Floor", self)
        # open_file.setShortcut('Ctrl+R')
        noise_floor.setStatusTip("Noise Floor.")
        self.connect(noise_floor, QtCore.SIGNAL("triggered()"), self.noise_floor)

        protein_track = QtGui.QAction("iSCAT Protein", self)
        protein_track.setStatusTip("iSCAT Protein")
        self.connect(protein_track, QtCore.SIGNAL("triggered()"), self.protein_projects)

        iPSF_model = QtGui.QAction("iPSF model", self)
        iPSF_model.setStatusTip("iPSF model")
        self.connect(iPSF_model, QtCore.SIGNAL("triggered()"), self.iPSF_projects)

        analyze_menu.addAction(noise_floor)
        analyze_menu.addAction(protein_track)
        analyze_menu.addAction(iPSF_model)

        # --------Analyze.--------

        # --------Help-----------
        help_menu = menubar.addMenu("&Help")

        help = QtGui.QAction("Help", self)
        # open_file.setShortcut('Ctrl+R')
        help.setStatusTip("Help.")
        self.connect(help, QtCore.SIGNAL("triggered()"), self.help)

        tutorials = QtGui.QAction("tutorials", self)
        # open_file.setShortcut('Ctrl+R')
        tutorials.setStatusTip("tutorials.")
        self.connect(tutorials, QtCore.SIGNAL("triggered()"), self.tutorials)

        about = QtGui.QAction("About", self)
        # open_file.setShortcut('Ctrl+R')
        about.setStatusTip("About.")
        self.connect(about, QtCore.SIGNAL("triggered()"), self.about)

        help_menu.addAction(help)
        # help_menu.addAction(tutorials)
        help_menu.addAction(about)
        # --------About-----------

        # --------Information Box--------
        save_txt = QtGui.QAction("Report", self)
        # save_txt.setShortcut('Ctrl+Q')
        save_txt.setStatusTip("save history")
        save_txt.triggered.connect(self.save_text_history)
        toolbar_save = self.addToolBar("Save")
        toolbar_save.addAction(save_txt)

        clear_txt = QtGui.QAction("Clear", self)
        # exitAction.setShortcut('Ctrl+Q')
        clear_txt.setStatusTip("clear Information Box")
        clear_txt.triggered.connect(self.clear_plain_text)

        toolbar_clear = self.addToolBar("Clear")
        toolbar_clear.addAction(clear_txt)

        exitAction = QtGui.QAction("Exit", self)
        # exitAction.setShortcut('Ctrl+Q')
        exitAction.setStatusTip("Exit application")
        exitAction.triggered.connect(self.closeEvent)
        fileMenu.addAction(exitAction)
        toolbar = self.addToolBar("Exit")
        toolbar.addAction(exitAction)
        # --------Information Box--------

        self.statusbar = self.statusBar()
        self.progressBar = QtWidgets.QProgressBar(self)
        self.statusBar().addPermanentWidget(self.progressBar)
        self.progressBar.setGeometry(30, 40, 200, 25)
        self.progressBar.setValue(0)
        self.update_progressBar = ProgressBar(self.progressBar)

        self.show()

    def __del__(self):
        print("Destructor called, Employee deleted.")

    def closeEvent(self, event):
        QtCore.QCoreApplication.instance().quit()
        print("closing PlaySetting")

    @QtCore.Slot()
    def set_new_text(self, input_text):
        if type(input_text) is str:
            self.textEdit.append(input_text)

    @QtCore.Slot()
    def set_plain_text(self, input_text):
        if type(input_text) is str:
            self.textEdit.insertPlainText(input_text)

    def clear_plain_text(self):
        self.textEdit.clear()

    def save_text_history(self):
        self.file_path = False
        self.file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "project_report", QtCore.QDir.currentPath()
        )
        timestr = time.strftime("%Y%m%d-%H%M%S")

        with open(self.file_path + "_" + timestr + ".txt", "w") as outfile:
            outfile.write(str(self.textEdit.toPlainText()))

    def spectrum_init(self):
        try:
            self.FFT2D_GUI_wrapper.spectrum_input(self.reading.original_video)
            self.FFT2D_GUI_wrapper.spectrum()

        except AttributeError:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("input video does not find!")
            self.msg_box.exec_()

    def analysis_wrapper(self):
        self.analysis_videos = Analysis()
        self.analysis_videos.update_output.connect(
            partial(self.updata_input_video, label="Video_analysis")
        )

    def analysis_wrapper_2(self):
        self.analysis_constant = AnalysisConstant()
        self.analysis_constant.update_output.connect(
            partial(self.updata_input_video, label="Video_analysis_constant")
        )

    def cpu_setting(self):
        self.processor_setting = CPU_setting_wrapper()

    def available_video_in_memory(self):
        self.vid_in_memory = VideoInMemory(self.list_available_video)
        self.vid_in_memory.display_trigger.connect(self.video_display)

    def video_loading_wrapper(self):
        self.reading = Reading()
        self.reading.update_output.connect(partial(self.updata_input_video, label="loading"))
        self.reading.read_video()

    def image2video_loading_wrapper(self):
        self.reading = Reading()
        self.reading.update_output.connect(partial(self.updata_input_video, label="loading"))
        self.reading.im2video()

    def tiff_video_loading_wrapper(self):
        self.reading = Reading()
        self.reading.update_output.connect(partial(self.updata_input_video, label="loading"))
        self.reading.tiff_video()

    def noise_floor(self):
        self.noise_floor_ = Noise_Floor(video=self.original_video)

    def protein_projects(self):
        self.protein_gui = ProteinTabs(
            video_in=self.original_video,
            batch_size=None,
            object_update_progressBar=self.update_progressBar,
        )
        self.protein_gui.new_update_DRA_video.connect(
            partial(self.updata_input_video, label="DRA", flag_DRA=True)
        )
        self.protein_gui.show()

    def iPSF_projects(self):
        self.iPSF_gui = GUI_iPSF()
        self.iPSF_gui.display_trigger.connect(self.video_display)
        self.iPSF_gui.show()

    def updata_input_video(self, data_in, label, flag_DRA=False):
        if flag_DRA:
            self.dra_video = data_in[0]
            self.batch_size = data_in[3]
            self.list_available_video["DRA_video"] = True
            status_line_info = None
        else:
            try:
                status_line_info = data_in[3]
            except:
                status_line_info = None
            self.original_video = data_in[0]
            self.list_available_video["original_video"] = True
        title = data_in[1]
        file_name = data_in[2]

        self.video_in_memory_flag[label] = True
        self.set_new_text("**** " + str(datetime.datetime.now()) + " ****")
        self.set_new_text(file_name)
        self.set_new_text(label + " " + title + " shape:" + str(self.original_video.shape))

        if status_line_info is not None and status_line_info["status_line_position"] != "":
            self.set_new_text(
                "---Status line detected in " + status_line_info["status_line_position"] + "---"
            )
        elif status_line_info is not None and status_line_info["status_line_position"] == "":
            self.set_new_text("---Status line does not detect---")

    def video_display(self, data_in):
        if "original_video" == data_in:
            self.visualization_ = Visulization_localization()
            self.visualization_.new_display(
                self.original_video,
                self.original_video,
                object=None,
                title="RAW",
                mask_status=False,
            )

        elif "DRA_video" == data_in:
            self.visualization_ = Visulization_localization()
            self.visualization_.new_display(
                self.dra_video, self.dra_video, object=None, title="DRA", mask_status=False
            )

        elif "iPSF_Model_I" == data_in[0] or "iPSF_Model_II" == data_in[0]:
            self.visualization_ = Visulization_localization()
            self.visualization_.new_display(
                data_in[1],
                data_in[1],
                object=self.iPSF_gui,
                title="iPSF Model",
                mask_status=False,
            )

    def tutorials(self):
        subprocess.run("python -m piscat.Tutorials", shell=True)

    def about(self):
        self.msg_box = QtWidgets.QMessageBox()
        self.msg_box.setWindowTitle("About PiSCAT")
        self.msg_box.setText("PiSCAT version %s" % version)
        self.msg_box.exec_()

    def help(self):
        self.help_windows = Help()

    def open_dialog_box(self):
        full_path, _ = QtWidgets.QFileDialog.getOpenFileNames(
            self, "Select Multi File", os.getcwd()
        )

        # Store files in list
        self.file_list = []
        [self.file_list.append(path) for path in full_path]

    def batch_analysis(self):
        self.batch_analysis_gui = BatchAnalysis(object_update_progressBar=self.update_progressBar)
        self.batch_analysis_gui.show()


def main():
    if sys.argv[0][-4:] == ".exe":
        sys.frozen = True

    multiprocessing.freeze_support()
    app = QtWidgets.QApplication(sys.argv)
    mpl_icon = pkg_resources.resource_filename("piscat.GUI.icons", "mpl.png")
    app.setWindowIcon(QtGui.QIcon(mpl_icon))

    ex = PiSCAT_GUI()

    print("Starting PiSCAT version %s" % version)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import numpy as np
from tqdm.autonotebook import tqdm


class TemporalFilter:
    def __init__(self, video, batchSize):
        """Filters to be applied to temporal features are included in this
        class.

        Parameters
        ----------
        video: NDArray
            Input video.

        batchSize: int
            Batch size that is used for DRA.

        """
        self.video = video
        self.batchSize = batchSize

    def filter_tarj_base_length(self, df_PSFs, threshold_min, threshold_max):
        """This function removes the particle from data frames that have a
        temporal length smaller and bigger than threshold_min values.

        Parameters
        ----------
        df_PSFs: pandas dataframe
            The data frame contains PSFs locations and ID (x, y, frame, sigma,
            particle, ...)

        threshold_min: int
            The minimum acceptable temporal length of one particle.

        threshold_max: int
            The maximum acceptable temporal length of one particle.

        Returns
        -------
        particles: pandas dataframe
            The filter data frame (x, y, frame, sigma, particle, ...)

        his_all_particles: list
            List that shows the statistics of particles length.

        """
        if df_PSFs.shape[0] == 0 or df_PSFs is None:
            raise ValueError("---data frames is empty!---")

        his_all_particles = df_PSFs["particle"].value_counts()
        temp_0 = his_all_particles.where(his_all_particles >= threshold_min)
        temp = his_all_particles.where(temp_0 <= threshold_max)
        select_particles = temp[~temp.isnull()]
        index_particles = select_particles.index
        particles = df_PSFs.loc[df_PSFs["particle"].isin(index_particles)]
        return particles, his_all_particles

    def v_trajectory(self, df_PSFs, threshold_min, threshold_max):
        """This function extract v-shape of the particle that has a temporal
        length between threshold_min and threshold_min max values.

        Parameters
        ----------
        df_PSFs: pandas dataframe
            The data frame contains PSFs locations and ID (x, y, frame, sigma, particle, ...).

        threshold_min: int
            The minimum acceptable temporal length of one particle.

        threshold_max: int
            The maximum acceptable temporal length of one particle.

        Returns
        -------
        all_trajectories: List of list
            Returns list of extracted data (i.e [List of list])

            | [intensity_horizontal, intensity_vertical,
               particle_center_intensity, particle_center_intensity_follow,
               particle_frame, particle_sigma, particle_X, particle_Y,
               particle_ID, optional(fit_intensity, fit_x, fit_y, fit_X_sigma,
               fit_Y_sigma, fit_Bias, fit_intensity_error, fit_x_error,
               fit_y_error, fit_X_sigma_error, fit_Y_sigma_error,
               fit_Bias_error)]

        particles: pandas dataframe
             The dataframe after using temporal filter (x, y, frame, sigma, particle, ...)

        his_all_particles: list
            List that shows the statistics of particles length.

        """
        if df_PSFs.shape[0] == 0 or df_PSFs is None:
            raise ValueError("---data frames is empty!---")

        particles, his_all_particles = self.filter_tarj_base_length(
            df_PSFs=df_PSFs, threshold_min=threshold_min, threshold_max=threshold_max
        )
        all_trajectories = self.v_profile(df_PSFs=particles, window_size=self.batchSize)
        return all_trajectories, particles, his_all_particles

    def v_profile(self, df_PSFs, window_size=2000):
        """The V-Shape trajectories and extended version are calculated.

        Parameters
        ----------
        window_size: int
            The maximum number of the frames that follow the V-Shape contrast
            from both sides.

        Returns
        -------
        all_trajectories: List of list
            Returns array contains the following information for each particle.

            | [intensity_horizontal, intensity_vertical,
               particle_center_intensity, particle_center_intensity_follow,
               particle_frame, particle_sigma, particle_X, particle_Y,
               particle_ID, optional(fit_intensity, fit_x, fit_y, fit_X_sigma,
               fit_Y_sigma, fit_Bias, fit_intensity_error, fit_x_error,
               fit_y_error, fit_X_sigma_error, fit_Y_sigma_error,
               fit_Bias_error)]

        """

        all_trajectories = []
        intensity_trajectories_dic = dict()
        index_particles = df_PSFs["particle"].unique().tolist()

        print("\nstart V_trajectories without parallel loop--->", end=" ")
        window_size = 2 * window_size

        for j_ in tqdm(index_particles):
            particle = df_PSFs.loc[df_PSFs["particle"] == j_]
            particle_ID = particle["particle"].tolist()
            particle_X = particle["x"].tolist()
            particle_Y = particle["y"].tolist()
            particle_frame = particle["frame"].tolist()
            particle_center_intensity = particle["center_intensity"].tolist()
            particle_sigma = particle["sigma"].tolist()

            if "Fit_Amplitude" in particle.keys():
                fit_intensity = particle["Fit_Amplitude"].tolist()
                fit_x = particle["Fit_X-Center"].tolist()
                fit_y = particle["Fit_Y-Center"].tolist()
                fit_X_sigma = particle["Fit_X-Sigma"].tolist()
                fit_Y_sigma = particle["Fit_Y-Sigma"].tolist()
                fit_Bias = particle["Fit_Bias"].tolist()

                fit_intensity_error = particle["Fit_errors_Amplitude"].tolist()
                fit_x_error = particle["Fit_errors_X-Center"].tolist()
                fit_y_error = particle["Fit_errors_Y-Center"].tolist()
                fit_X_sigma_error = particle["Fit_errors_X-Sigma"].tolist()
                fit_Y_sigma_error = particle["Fit_errors_Y-Sigma"].tolist()
                fit_Bias_error = particle["Fit_errors_Bias"].tolist()

            intensity_vertical = []
            intensity_horizontal = []
            for frame_num, y, x, sigma in zip(
                particle_frame, particle_Y, particle_X, particle_sigma
            ):
                len_profile = int(5 * sigma)
                min_v = np.max([y - len_profile, 0])
                max_v = np.min([y + len_profile, self.video.shape[1]])

                intensity_vertical.append(
                    self.video[int(frame_num), int(min_v) : int(max_v), int(x)]
                )

                min_h = np.max([x - len_profile, 0])
                max_h = np.min([x + len_profile, self.video.shape[1]])

                intensity_horizontal.append(
                    self.video[int(frame_num), int(y), int(min_h) : int(max_h)]
                )

            first_frame = int(particle_frame[0])
            x_first_frame = int(particle_X[0])
            y_first_frame = int(particle_Y[0])

            last_frame = int(particle_frame[-1])
            x_last_frame = int(particle_X[-1])
            y_last_frame = int(particle_Y[-1])

            start_frame = np.max([0, first_frame - window_size])
            end_frame = np.min([self.video.shape[0], last_frame + window_size])
            particle_center_intensity = []
            for f_, x_, y_ in zip(particle_frame, particle_X, particle_Y):
                particle_center_intensity.append(self.video[int(f_), int(y_), int(x_)])

            particle_center_intensity_follow = self.video[start_frame:end_frame, y_first_frame, x_first_frame]
            particle_frame_follow = list(range(int(start_frame), int(end_frame)))


            # particle_center_intensity_follow_backward = self.video[
            #     int(start_frame) : int(first_frame), y_first_frame, x_first_frame
            # ]
            # particle_center_intensity_follow_forward = self.video[
            #     int(last_frame) : int(end_frame), y_last_frame, x_last_frame
            # ]

            # particle_center_intensity_follow = np.concatenate(
            #     (
            #         particle_center_intensity_follow_backward,
            #         particle_center_intensity,
            #         particle_center_intensity_follow_forward,
            #     ),
            #     axis=0,
            # )
            trajectories = []
            trajectories.append(intensity_horizontal)
            trajectories.append(intensity_vertical)
            trajectories.append(particle_center_intensity)
            trajectories.append(particle_center_intensity_follow)
            trajectories.append(particle_frame)
            trajectories.append(particle_frame_follow)
            trajectories.append(particle_sigma)
            trajectories.append(particle_X)
            trajectories.append(particle_Y)
            trajectories.append(particle_ID)
            if "Fit_Amplitude" in particle.keys():
                trajectories.append(fit_intensity)
                trajectories.append(fit_x)
                trajectories.append(fit_y)
                trajectories.append(fit_X_sigma)
                trajectories.append(fit_Y_sigma)
                trajectories.append(fit_Bias)
                trajectories.append(fit_intensity_error)
                trajectories.append(fit_x_error)
                trajectories.append(fit_y_error)
                trajectories.append(fit_X_sigma_error)
                trajectories.append(fit_Y_sigma_error)
                trajectories.append(fit_Bias_error)

            all_trajectories.append(trajectories)
        print("Done")

        return all_trajectories



import os
import pickle


def dump_array(file_name, arr):
    if os.path.exists(file_name):
        raise Exception("File already exists!")

    with open(file_name, "wb") as f:
        pickle.dump(arr, f)



import os
import shutil
import subprocess
import sys
from os.path import exists
from pathlib import Path


def parse_args():
    """Parses the command line arguments.

    The first command line argument is the directory where to copy
    the tutorials.

    If no command line argument is given, assume working directory
    as target, but ask for user confirmation.

    Raises
    ------
    Exception
        if command line arguments are invalid (e.g. more than one
        argument)

    """

    if len(sys.argv) <= 1:
        if confirm(
            "No location given. Shall the tutorials be copied to "
            "the current working directory? [y/n] "
        ):
            target_root = Path.cwd()
        else:
            raise Exception("Exiting.")
    elif len(sys.argv) > 2:
        raise Exception(
            "ERROR! Too many arguments!\n\n" "USAGE: python -m piscat.Tutorials TARGET_DIRECTORY"
        )
    else:
        target_root = Path(sys.argv[1])

    return target_root


def find_jupyter():
    """Try to find the Jupyter executable on the system.

    Returns
    -------
    jupyter: str
        The command to start jupyter

    Raises
    ------
    Exception
        if no jupyter executable is found on the system.
    """

    jupyter = shutil.which("jupyter")
    if jupyter:
        jupyter += " notebook"

    if not jupyter:
        jupyter = shutil.which("jupyter-notebook")

    if not jupyter:
        raise Exception("Jupyter not found on your system. Please start it manually.")

    return jupyter


def start_server(root, jupyter):
    """Start up Jupyter server.

    Parameters
    ----------
    root: pathlib.Path
        The directory in which to start the server.

    jupyter: str
        The console command to start the jupyter server.

    """
    os.chdir(root)
    subprocess.run("jupyter nbextension enable --py widgetsnbextension", shell=True)
    subprocess.run(jupyter, shell=True)


def confirm(prompt):
    """Get user confirmation from standard input.

    Parameters
    ----------
    prompt: str
        The message to show.

    Returns
    -------
    True or false depending on whether the user has answered with yes or no.
    """
    answer = ""
    while answer.lower() != "y" and answer.lower() != "n":
        answer = input(prompt)
    return answer == "y"


def copy_tutorials(target_root):
    """Copies the PiSCAT tutorials to a given directory.

    Parameters
    ----------
    target_root: pathlib.Path
        Where to copy the directories. If already existing, files
        are not copied to prevent accidental overwriting.

    Returns
    -------
    target: pathlib.Path
        The path to the directory with the tutorials.
    """

    target = target_root / "Tutorials/JupyterFiles"
    src = Path(__file__).parent / "JupyterFiles"

    if exists(target):
        Warning("Warning! Target directory %s already exists." % str(target))
    else:
        print("Copying tutorials to " + str(target))
        shutil.copytree(src, target)

    return target


def start_tutorials():
    """Copy the PiSCAT tutorials to the target directory and start Jupyter server

    The target directory is the first argument of sys.argv.
    """

    target_root = parse_args()
    root = copy_tutorials(target_root)
    jupyter = find_jupyter()
    start_server(root, jupyter)



from PySide6 import QtCore, QtWidgets


#############Define MyWindow Class Here ############
class SaveVideo(QtWidgets.QMainWindow):
    signal_save_Done = QtCore.Signal()

    ##-----------------------------------------
    def __init__(self, parent=None):
        super().__init__(parent)
        self.width_size = None
        self.height_size = None

        self.width_size_s = None
        self.width_size_e = None

        self.height_size_s = None
        self.height_size_e = None

        self.frame_s = None
        self.frame_e = None
        self.frame_jump = None

        self.set_bit_order = None
        self.bin_type = None
        self.video_type = None

        self.flag_display = False

        self.image_format = {
            "int_8": "i1",
            "int_16": "i2",
            "int_32": "i4",
            "int_64": "i8",
            "uint_8": "u1",
            "uint_16": "u2",
            "uint_32": "u4",
            "uint_64": "u8",
            "float_16": "f2",
            "float_32": "f4",
            "float_64": "f8",
        }

        self.raw_data_update_flag = True

        self.window = QtWidgets.QWidget()

        self.combo_video = QtWidgets.QComboBox(self)
        self.combo_video.addItem("-Select the video type-")
        self.combo_video.addItem("RAW")
        self.combo_video.addItem("MP4")
        self.combo_video.addItem("GIF")

        self.combo_bin = QtWidgets.QComboBox(self)
        self.combo_bin.addItem("-Select the binary type-")
        self.combo_bin.addItem("int_8")
        self.combo_bin.addItem("int_16")
        self.combo_bin.addItem("int_32")
        self.combo_bin.addItem("int_64")

        self.combo_bin.addItem("uint_8")
        self.combo_bin.addItem("uint_16")
        self.combo_bin.addItem("uint_32")
        self.combo_bin.addItem("uint_64")

        self.combo_bin.addItem("float_16")
        self.combo_bin.addItem("float_32")
        self.combo_bin.addItem("float_64")
        self.combo_bin.currentIndexChanged.connect(self.on_select)

        self.ok = QtWidgets.QPushButton("Ok")
        self.ok.setAutoDefault(False)
        self.ok.setFixedWidth(100)
        self.ok.clicked.connect(self.do_update)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)
        self.grid.addWidget(self.ok, 6, 0)

        self.setWindowTitle("Saving video")
        self.setStyleSheet("QMainWindow{background-color: darkgray;}")
        self.window.setLayout(self.grid)

    def createFirstExclusiveGroup(self):
        self.groupBox_cropping = QtWidgets.QGroupBox("Cast to a specified bin_type")
        self.groupBox_cropping.setCheckable(True)
        self.groupBox_cropping.setChecked(True)

        # groupBox.isChecked(False)
        self.grid1 = QtWidgets.QGridLayout()
        self.grid1.addWidget(self.combo_video, 0, 0)
        self.grid1.addWidget(self.combo_bin, 0, 1)
        self.groupBox_cropping.setLayout(self.grid1)

        return self.groupBox_cropping

    def do_update(self):
        if self.ok.clicked:
            self.on_select()
            if self.video_type is not None:
                if self.video_type == "RAW":
                    if self.bin_type is None:
                        self.signal_save_Done.emit()
                    else:
                        self.signal_save_Done.emit()
                    self.raw_data_update_flag = False
                elif self.video_type == "MP4":
                    self.signal_save_Done.emit()
                elif self.video_type == "GIF":
                    self.signal_save_Done.emit()

    def on_select(self):
        if self.combo_bin.currentText() == "-Select the binary type-":
            self.bin_type = None
        else:
            self.bin_type = self.image_format[self.combo_bin.currentText()]

        if self.combo_video.currentText() == "-Select the video type-":
            self.video_type = None
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Please select type of video writer!")
            self.msg_box.exec_()
        else:
            self.video_type = self.combo_video.currentText()

    def closeEvent(self, **kwargs):
        QtCore.QCoreApplication.instance().quit()



import os
import time

import imageio
from tqdm.autonotebook import tqdm


def write_binary(dir_path, file_name, data, type="original"):
    """
    This function writes video as a binary.

    Parameters
    ----------
    dir_path: str
        Path of the directory that video save on it.

    file_name: str
       Name of the save file.

    data: NDArray
        Video with numpy format.

    type: str or bin_type
        The video bin type is not changed by 'original,' but the user can
        convert it (e.g. float --> int16).

    Returns
    -------
    The path to the new folder that was created to save the video is returned.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    try:
        dr_mk = os.path.join(dir_path, timestr)
        os.mkdir(dr_mk)
        print("Directory ", timestr, " Created ")
    except FileExistsError:
        print("Directory ", timestr, " already exists")

    save_path = os.path.join(dir_path, timestr, file_name)
    save_path_ = os.path.join(dir_path, timestr)

    if type == "original":
        data = data
    else:
        data = data.astype(type)

    data = data.copy(order="C")
    with open(save_path, "wb") as outfile:
        outfile.write(data)

    return save_path_


def write_MP4(dir_path, file_name, data, jump=0):
    """
    This function writes video as a MP4.

    Parameters
    ----------
    dir_path: str
        Path of the directory that video save on it.

    file_name: str
       Name of the save file.

    data: NDArray
        Video with numpy format.

    jump: int
        Define stride between frames.

    Returns
    -------
    The path to the new folder that was created to save the video is returned.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    try:
        dr_mk = os.path.join(dir_path, timestr)
        os.mkdir(dr_mk)
        print("Directory ", timestr, " Created ")
    except FileExistsError:
        print("Directory ", timestr, " already exists")

    save_path = os.path.join(dir_path, timestr, file_name)
    save_path_ = os.path.join(dir_path, timestr)

    data = data.copy(order="C")
    image_ = []
    for frame_number in tqdm(range(0, data.shape[0] - jump, jump)):
        image_.append(data[frame_number, ...])

    imageio.mimsave(save_path, image_, format="MP4")

    return save_path_


def write_GIF(dir_path, file_name, data, jump=0):
    """
    This function writes video as a GIF.

    Parameters
    ----------
    dir_path: str
        Path of the directory that video save on it.

    file_name: str
       Name of the save file

    data: NDArray
        Video with numpy format.

    jump: int
        Define stride between frames.

    Returns
    -------
    The path to the new folder that was created to save the video is returned.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    try:
        dr_mk = os.path.join(dir_path, timestr)
        os.mkdir(dr_mk)
        print("Directory ", timestr, " Created ")
    except FileExistsError:
        print("Directory ", timestr, " already exists")

    save_path = os.path.join(dir_path, timestr, file_name)
    save_path_ = os.path.join(dir_path, timestr)

    data = data.copy(order="C")
    image_ = []
    for frame_number in tqdm(range(0, data.shape[0] - jump, jump)):
        image_.append(data[frame_number, ...])

    imageio.mimsave(save_path, image_, format="GIF")
    return save_path_



import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QRunnable, Signal
from PySide6.QtGui import QPen, Qt
from scipy.ndimage import median_filter
from scipy.signal import savgol_filter

from piscat.GUI.Visualization.updating_plots import (
    UpdatingPlotsPyqtGraphSpatial,
    UpdatingPlotsPyqtGraphTemporal,
)
from piscat.Localization.directional_intensity import DirectionalIntensity
from piscat.Preproccessing.normalization import Normalization
from piscat.Visualization.contrast_adjustment import ContrastAdjustment


class SliceView(QtWidgets.QGraphicsView, QRunnable):
    photoClicked = QtCore.Signal(tuple)
    frameClicked = QtCore.Signal(int)
    pixmapClicked = QtCore.Signal(QtCore.QPoint)
    cursorMove = QtCore.Signal(object)
    annotation_s = QtCore.Signal()
    annotation_t = QtCore.Signal(object)
    finished = Signal()

    x0 = 0
    y0 = 0
    x1 = 0
    y1 = 0
    flag_paint = False

    def __init__(self, input_video, video_original, stride=1, mask=False, *args, **kwargs):
        self.scene = QtWidgets.QGraphicsScene()
        super(SliceView, self).__init__(self.scene, *args, **kwargs)

        self.updatePlot_spatial = UpdatingPlotsPyqtGraphSpatial()
        self.updatePlot_temporal = UpdatingPlotsPyqtGraphTemporal()

        self.RAW_Video = video_original

        if video_original.ndim == 3:
            self.video_width = video_original.shape[1]
            self.video_height = video_original.shape[2]
        elif video_original.ndim == 2:
            self.video_width = video_original.shape[0]
            self.video_height = video_original.shape[1]

        self.input_video = input_video
        self.con_adj = ContrastAdjustment(input_video)
        self.current_pixmap = None
        self.current_mask_pixmap = None
        self.maskArray = None
        self.mask_is_set = False
        self.live_spatial_paint_Flag = False
        self.live_temporal_paint_Flag = False
        self.medianFilterFlag = False
        self.ori_X0 = 0
        self.ori_Y0 = 0
        self.ori_X1 = 0
        self.ori_Y1 = 0
        self.pixel_value = 0
        self.slice_num = 0
        self.marker_size = 5

        self._zoom = 0
        self._empty = True

        self.annotation_s.connect(self.active_spatial_live_paint)
        self.annotation_t.connect(self.active_temporal_live_paint)

        self.setBackgroundBrush(QtGui.QColor(0, 0, 0))
        self.setScene(self.scene)

        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)

        if mask is True:
            self.initialze_mask()

    def __del__(self):
        print("Destructor called, Employee deleted.")

    def enterEvent(self, event):
        self.setMouseTracking(True)

    def leaveEvent(self, event):
        self.setMouseTracking(False)

    def pollCursor(self):
        pos = QtGui.QCursor.pos()
        if pos != self.cursor:
            self.cursor = pos
            self.cursorMove.emit(pos)

    def handleCursorMove(self, pos):
        print(pos)

    def hasPhoto(self):
        return not self._empty

    def toggleDragMode(self):
        if self.dragMode() == QtWidgets.QGraphicsView.ScrollHandDrag:
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
        elif not self._photo.pixmap().isNull():
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)

    def mousePressEvent(self, event):
        if self._photo.isUnderMouse():
            sp = self.mapToScene(event.pos())
            lp = self._photo.mapFromScene(sp).toPoint()

            self.x0 = lp.x()
            self.y0 = lp.y()
            self.flag_paint = True

            self.ori_X0 = int(
                (self.RAW_Video.shape[1] / self.pixmap_video.size().toTuple()[0]) * self.x0
            )
            self.ori_Y0 = int(
                (self.RAW_Video.shape[2] / self.pixmap_video.size().toTuple()[1]) * self.y0
            )

            self.frameClicked.emit(self.slice_num)
            self.photoClicked.emit((self.ori_X0, self.ori_Y0))
            self.paint_point()

        super(SliceView, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.flag_paint = False

    def mouseMoveEvent(self, event):
        if self.flag_paint:
            if self._photo.isUnderMouse():
                sp = self.mapToScene(event.pos())
                lp = self._photo.mapFromScene(sp).toPoint()

                self.x1 = lp.x()
                self.y1 = lp.y()

                self.ori_X1 = int(
                    (self.RAW_Video.shape[1] / self.pixmap_video.size().toTuple()[0]) * self.x1
                )
                self.ori_Y1 = int(
                    (self.RAW_Video.shape[2] / self.pixmap_video.size().toTuple()[1]) * self.y1
                )

                self.paint_line()

    def initialze_mask(self):
        if self.input_video.shape[1] >= self.input_video.shape[2]:
            self.mask_qimage = QtGui.QImage(
                QtCore.QSize(self.input_video.shape[2], self.input_video.shape[1]),
                QtGui.QImage.Format_ARGB32,
            )

            self.mask_qimage.fill(QtGui.QColor(250, 240, 0, 0))
            self.mask_qimage_view = np.frombuffer(
                self.mask_qimage.bits(), dtype=np.uint8
            ).reshape(self.input_video.shape[1], self.input_video.shape[2], 4)
        else:
            self.mask_qimage = QtGui.QImage(
                QtCore.QSize(self.input_video.shape[2], self.input_video.shape[1]),
                QtGui.QImage.Format_ARGB32,
            )

            self.mask_qimage.fill(QtGui.QColor(250, 240, 0, 0))
            self.mask_qimage_view = np.frombuffer(
                self.mask_qimage.bits(), dtype=np.uint8
            ).reshape(self.input_video.shape[1], self.input_video.shape[2], 4)

    def create_mask_pixmap(self, input_mask_slide):
        self.mask_qimage_view[..., 3] = input_mask_slide * int(1 * 255)
        self.current_mask_pixmap = QtGui.QPixmap.fromImage(self.mask_qimage)
        return self.current_mask_pixmap

    def create_pixmap(self, input_file):
        if isinstance(input_file, str):
            self.input_file = input_file

        else:
            input_file = Normalization(video=input_file).normalized_image_specific()
            if self.parent().parent().alpha is None:
                input_file = self.con_adj.auto_pixelTransforms(input_file)
                self.input_file = input_file
            else:
                alpha = self.parent().parent().alpha
                beta = self.parent().parent().beta
                min_intensity = self.parent().parent().min_intensity
                max_intensity = self.parent().parent().max_intensity
                input_file = self.con_adj.pixel_transforms(
                    input_file, alpha, beta, min_intensity, max_intensity
                )
                self.input_file = input_file

        if self.medianFilterFlag:
            input_file = median_filter(input_file, 3)

        if self.parent().parent().title != "PNG":
            if input_file.shape[0] >= input_file.shape[1]:
                widthStep = input_file.shape[1]
                img = QtGui.QImage(
                    input_file,
                    input_file.shape[1],
                    input_file.shape[0],
                    widthStep,
                    QtGui.QImage.Format_Indexed8,
                )
                img.scaledToWidth(input_file.shape[0])
            else:
                widthStep = input_file.shape[1]
                img = QtGui.QImage(
                    input_file,
                    input_file.shape[1],
                    input_file.shape[0],
                    widthStep,
                    QtGui.QImage.Format_Indexed8,
                )

            newimg = img.convertToFormat(QtGui.QImage.Format.Format_RGB888)
            self.current_pixmap = QtGui.QPixmap(newimg)
            self.update_slice(self.current_pixmap)
        elif self.parent().parent().title == "PNG":
            self.current_pixmap = QtGui.QPixmap(input_file)

        return self.current_pixmap

    def update_slice(self, input_pixmap):
        self.t = input_pixmap
        self._zoom = 0
        width = np.max([self.video_width, 400])
        height = np.max([self.video_height, 400])

        self.pixmap_video = input_pixmap.scaled(
            QtCore.QSize(width, height), QtCore.Qt.KeepAspectRatio
        )
        self._photo = self.scene.addPixmap(self.pixmap_video)
        if self.live_spatial_paint_Flag:
            self.paint_line()
        if self.live_temporal_paint_Flag:
            self.paint_point()

    def update_overlay(self, input_mask_pixmap, first_flag=True):
        width = np.max([self.video_width, 400])
        height = np.max([self.video_height, 400])
        self.scaled_mask_pixmap = input_mask_pixmap.scaled(
            QtCore.QSize(width, height), QtCore.Qt.KeepAspectRatio
        )

        self._photo.setPixmap(self.scaled_mask_pixmap)
        self._photo_mask = self.scene.addPixmap(self.scaled_mask_pixmap)

        if self.parent().parent().title != "PNG" and first_flag is True:
            self.parent().parent().slice_slider.setValue(self.slice_num)
            self.current_pixmap = self.create_pixmap(self.input_video[self.slice_num, :, :])
            self.update_slice(self.current_pixmap)
            if self.mask_is_set is True:
                self.current_mask_pixmap = self.create_mask_pixmap(
                    self.maskArray[self.slice_num, :, :]
                )
                self.update_overlay(self.current_mask_pixmap, first_flag=False)

    def wheelEvent(self, event):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                factor = 1.25
                self._zoom += 1
            elif event.angleDelta().y() < 0:
                factor = 0.75
                self._zoom -= 1
            else:
                factor = 0.8
                self._zoom -= 1

            if self._zoom > 0:
                self.scale(factor, factor)
            elif self._zoom < 0:
                self.scale(factor, factor)
            else:
                self._zoom = 0

        else:
            if self.parent().parent().title == "PNG":
                pass
            elif self.parent().parent().title != "PNG":
                steps = event.angleDelta().y() // 120
                vector = steps and steps // abs(steps)  # 0, 1, or -1
                for _ in range(1, abs(steps) + 1):
                    self.slice_num -= vector
                    if self.slice_num >= 0 and self.slice_num <= self.input_video.shape[0] - 1:
                        self.parent().parent().slice_slider.setValue(self.slice_num)
                        self.current_pixmap = self.create_pixmap(
                            self.input_video[self.slice_num, :, :]
                        )
                        self.update_slice(self.current_pixmap)
                        if self.mask_is_set is True:
                            self.current_mask_pixmap = self.create_mask_pixmap(
                                self.maskArray[self.slice_num, :, :]
                            )
                            self.update_overlay(self.current_mask_pixmap)

    @QtCore.Slot()
    def active_spatial_live_paint(self):
        if self.live_spatial_paint_Flag is False:
            self.live_spatial_paint_Flag = True
        elif self.live_spatial_paint_Flag is True:
            self.live_spatial_paint_Flag = False
            try:
                self.scene.removeItem(self.Line)
                self.scene.update()
            except:
                pass

    @QtCore.Slot()
    def active_temporal_live_paint(self, setting=None):
        if self.live_temporal_paint_Flag is False:
            self.live_temporal_paint_Flag = True
        elif self.live_temporal_paint_Flag is True:
            self.live_temporal_paint_Flag = False

        try:
            self.scene.removeItem(self.point_)
            self.scene.update()
        except:
            pass

        self.read_temporal_pixel_values(setting=setting)

    def paint_line(self):
        pen = QPen(Qt.red)
        pen.setCosmetic(True)
        pen.setWidth(1)

        try:
            self.scene.removeItem(self.Line)
        except AttributeError:
            pass

        self.Line = self.scene.addLine(self.x0, self.y0, self.x1, self.y1, pen)
        self.scene.addItem(self.Line)
        self.scene.update()
        self.read_spatial_pixel_values()

    def paint_point(self):
        pen = QPen(Qt.red)
        pen.setCosmetic(True)
        pen.setWidth(1)

        try:
            self.scene.removeItem(self.Line)
        except AttributeError:
            pass

        self.point_ = self.scene.addEllipse(
            self.x0, self.y0, self.marker_size, self.marker_size, pen
        )
        self.scene.addItem(self.point_)
        self.scene.update()
        self.read_temporal_pixel_values(setting=None)

    def read_spatial_pixel_values(self):
        try:
            DI = DirectionalIntensity()
            radial_index = DI.interpolate_pixels_along_line(
                x0=int(self.ori_X0), y0=int(self.ori_Y0), x1=int(self.ori_X1), y1=int(self.ori_Y1)
            )
            frame = self.RAW_Video[self.slice_num, ...]

            pixels_ = []
            for idx_ in radial_index:
                pixels_.append(frame[int(idx_[1]), int(idx_[0])])

            x = list(range(0, len(pixels_)))
            self.updatePlot_spatial.data_line.setData(x, pixels_)  # Update the data.
            self.updatePlot_spatial.show()
        except:
            pass

    def read_temporal_pixel_values(self, setting=None):
        if setting is not None:
            self.win_size = setting[0]
            self.flag_smooth = setting[1]
            self.marker_size = setting[2]

        try:
            min_temporal_range = int(np.max((0, self.slice_num - self.win_size)))
            max_temporal_range = int(
                np.min((self.RAW_Video.shape[0], self.slice_num + self.win_size))
            )

            temporal_intensity = self.RAW_Video[
                min_temporal_range:max_temporal_range, self.ori_Y0, self.ori_X0
            ]
            # current_intensity = self.RAW_Video[self.slice_num, self.ori_Y0, self.ori_X0]

            if self.flag_smooth:
                smooth_window_size = round(0.05 * len(temporal_intensity))
                if (smooth_window_size % 2) == 0:
                    smooth_window_size = smooth_window_size + 1

                if self.win_size > 3:
                    temporal_intensity = savgol_filter(temporal_intensity, smooth_window_size, 3)

            x = list(range(min_temporal_range, max_temporal_range))
            idx_ = np.abs((min_temporal_range - self.slice_num))
            self.updatePlot_temporal.data_line.setData(x, temporal_intensity)  # Update the data.
            self.updatePlot_temporal.data_point.setData(
                [x[idx_]], [temporal_intensity[idx_]]
            )  # Update the data.
            self.updatePlot_temporal.show()
        except:
            pass



from .data_handling import protein_trajectories_list2dic
from .particle_linking import Linking
from .temporal_filtering import TemporalFilter

__all__ = [
    "protein_trajectories_list2dic",
    "Linking",
    "TemporalFilter",
]



import math

import numpy as np
from PySide6 import QtCore, QtWidgets

from piscat.GUI.Visualization.updating_plots import UpdatingPlots_Image
from piscat.iPSF_model import ImagingSetupParameters
from piscat.iPSF_model.ScatteredFieldDifferentialPhase import ScatteredFieldDifferentialPhase


class GUI_iPSF(QtWidgets.QWidget):
    display_trigger = QtCore.Signal(list)

    def __init__(self, parent=None):
        super(GUI_iPSF, self).__init__(parent)

        self.iPSFs_focalStack = None
        self.iPSFs_AxialStack = None

        self.empty_value_ImagingBox_flag = True
        self.empty_value_ModelBox_flag = True

        self.wavelength = QtWidgets.QLineEdit()
        self.wavelength.setToolTip("Wavelength of the light source in meters")
        self.wavelength.setPlaceholderText("Wavelength of the light source in meters")
        self.wavelength_label = QtWidgets.QLabel("wavelength:")
        self.wavelength.setFixedWidth(150)

        self.NA = QtWidgets.QLineEdit()
        self.NA.setToolTip("Numerical Aperture (NA) of the objective lens")
        self.NA.setPlaceholderText("Numerical Aperture (NA) of the objective lens")
        self.NA_label = QtWidgets.QLabel("Numerical Aperture (NA):")
        self.NA.setFixedWidth(150)

        self.ti0 = QtWidgets.QLineEdit()
        self.ti0.setToolTip("Thickness of the immersion oil, nominal value in meters")
        self.ti0.setPlaceholderText("Thickness of the immersion oil, nominal value in meters")
        self.ti0_label = QtWidgets.QLabel("Thickness:")
        self.ti0.setFixedWidth(150)

        self.ni0 = QtWidgets.QLineEdit()
        self.ni0.setToolTip("Refractive index of the immersion oil, nominal value")
        self.ni0.setPlaceholderText("Refractive index of the immersion oil, nominal value")
        self.ni0_label = QtWidgets.QLabel("Refractive index:")
        self.ni0.setFixedWidth(150)

        self.ni = QtWidgets.QLineEdit()
        self.ni.setToolTip("Refractive index of the immersion oil, experimental value")
        self.ni.setPlaceholderText("Refractive index of the immersion oil, experimental value")
        self.ni_label = QtWidgets.QLabel("Refractive index:")
        self.ni.setFixedWidth(150)

        self.tg0 = QtWidgets.QLineEdit()
        self.tg0.setToolTip("Thickness of the coverglass, nominal value in meters")
        self.tg0.setPlaceholderText("Thickness of the coverglass, nominal value in meters")
        self.tg0_label = QtWidgets.QLabel("Thickness:")
        self.tg0.setFixedWidth(150)

        self.tg = QtWidgets.QLineEdit()
        self.tg.setToolTip("Thickness of the coverglass, experimental value in meters")
        self.tg.setPlaceholderText("Thickness of the coverglass, experimental value in meters")
        self.tg_label = QtWidgets.QLabel("Thickness:")
        self.tg.setFixedWidth(150)

        self.ng0 = QtWidgets.QLineEdit()
        self.ng0.setToolTip("Refractive index of the coverglass, nominal value")
        self.ng0.setPlaceholderText("Refractive index of the coverglass, nominal value")
        self.ng0_label = QtWidgets.QLabel("Refractive index:")
        self.ng0.setFixedWidth(150)

        self.ng = QtWidgets.QLineEdit()
        self.ng.setToolTip("Refractive index of the coverglass, experimental value")
        self.ng.setPlaceholderText("Refractive index of the coverglass, experimental value")
        self.ng_label = QtWidgets.QLabel("Refractive index:")
        self.ng.setFixedWidth(150)

        self.ns = QtWidgets.QLineEdit()
        self.ns.setToolTip("Refractive index of the sample/medium")
        self.ns.setPlaceholderText("Refractive index of the sample/medium")
        self.ns_label = QtWidgets.QLabel("Refractive index:")
        self.ns.setFixedWidth(150)

        self.pixel_size_physical = QtWidgets.QLineEdit()
        self.pixel_size_physical.setToolTip("Physical size of the camera pixel in meters")
        self.pixel_size_physical.setPlaceholderText("Physical size of the camera pixel in meters")
        self.pixel_size_physical_label = QtWidgets.QLabel("Pixel size:")
        self.pixel_size_physical.setFixedWidth(150)

        self.pixel_size = QtWidgets.QLineEdit()
        self.pixel_size.setToolTip(
            (
                "Imaging pixel size in meters."
                "Related to the physical pixel size through the magnification of the setup."
            )
        )
        self.pixel_size.setPlaceholderText(
            (
                "Imaging pixel size in meters."
                "Related to the physical pixel size through the magnification of the setup."
            )
        )
        self.pixel_size_label = QtWidgets.QLabel("Pixel size:")
        self.pixel_size.setFixedWidth(150)

        self.s_range = QtWidgets.QLineEdit()
        self.s_range.setToolTip("The starting range across which the focus is swept")
        self.s_range.setPlaceholderText("focus start point in meter")
        self.s_range_label = QtWidgets.QLabel("focus is swept start point:")
        self.s_range.setFixedWidth(150)

        self.e_range = QtWidgets.QLineEdit()
        self.e_range.setToolTip("The ending range across which the focus is swept")
        self.e_range.setPlaceholderText("focus end point in meter")
        self.e_range_label = QtWidgets.QLabel("focus is swept end point:")
        self.e_range.setFixedWidth(150)

        self.step_range = QtWidgets.QLineEdit()
        self.step_range.setToolTip("The stride across which the focus is swept")
        self.step_range.setPlaceholderText("step in meter")
        self.step_range_label = QtWidgets.QLabel("focus is swept step:")
        self.step_range.setFixedWidth(150)

        self.s_range_particle = QtWidgets.QLineEdit()
        self.s_range_particle.setToolTip(
            "The starting axial range across which the particle is travelling"
        )
        self.s_range_particle.setPlaceholderText("particle start point in meter")
        self.s_range_label_particle = QtWidgets.QLabel("particle is swept start point:")
        self.s_range_particle.setFixedWidth(150)

        self.e_range_particle = QtWidgets.QLineEdit()
        self.e_range_particle.setToolTip(
            "The ending axial range across which the particle is travelling"
        )
        self.e_range_particle.setPlaceholderText("particle end point in meter")
        self.e_range_label_particle = QtWidgets.QLabel("particle is swept end point:")
        self.e_range_particle.setFixedWidth(150)

        self.step_range_particle = QtWidgets.QLineEdit()
        self.step_range_particle.setToolTip(
            "The stride of axial across which the particle is travelling"
        )
        self.step_range_particle.setPlaceholderText("step in meter")
        self.step_range_label_particle = QtWidgets.QLabel("particle is swept step:")
        self.step_range_particle.setFixedWidth(150)

        self.z_focus_particle = QtWidgets.QLineEdit()
        self.z_focus_particle.setToolTip("The position of the focal plane")
        self.z_focus_particle.setPlaceholderText("The position of the focal plane")
        self.z_focus_label_particle = QtWidgets.QLabel("The position of the focal plane:")
        self.z_focus_particle.setFixedWidth(150)

        self.position_particle3D = QtWidgets.QLineEdit()
        self.position_particle3D.setToolTip("The 3D position of the nanoparticle")
        self.position_particle3D.setPlaceholderText("[x, y, z]")
        self.position_particle3D_label = QtWidgets.QLabel("The 3D position of the nanoparticle:")
        self.position_particle3D.setFixedWidth(150)

        self.nx = QtWidgets.QLineEdit()
        self.nx.setToolTip("Number of lateral pixels over which the image is calculated")
        self.nx.setPlaceholderText("pixels")
        self.nx_label = QtWidgets.QLabel(
            "Number of lateral pixels over which the image is calculated :"
        )
        self.nx.setFixedWidth(150)

        self.nx_particle = QtWidgets.QLineEdit()
        self.nx_particle.setToolTip("Number of lateral pixels over which the image is calculated")
        self.nx_particle.setPlaceholderText("pixels")
        self.nx_label_particle = QtWidgets.QLabel(
            "Number of lateral pixels over which the image is calculated :"
        )
        self.nx_particle.setFixedWidth(150)

        self.r_ = QtWidgets.QLineEdit()
        self.r_.setToolTip("Number of lateral pixels over which the calculated image is cropped")
        self.r_.setPlaceholderText("pixels")
        self.r_label = QtWidgets.QLabel(
            "Number of lateral pixels over which the calculated image is cropped:"
        )
        self.r_.setFixedWidth(150)

        self.r_particle = QtWidgets.QLineEdit()
        self.r_particle.setToolTip(
            "Number of lateral pixels over which the calculated image is cropped"
        )
        self.r_particle.setPlaceholderText("pixels")
        self.r_label_particle = QtWidgets.QLabel(
            "Number of lateral pixels over which the calculated image is cropped:"
        )
        self.r_particle.setFixedWidth(150)

        self.generate_model = QtWidgets.QPushButton("Generate model")
        self.generate_model.clicked.connect(self.do_model)
        self.generate_model.setFixedWidth(150)

        self.display = QtWidgets.QPushButton("Display")
        self.display.clicked.connect(self.do_display)
        self.display.setFixedWidth(150)

        self.Initialization = QtWidgets.QPushButton("Initialization")
        self.Initialization.clicked.connect(self.do_Initialization)
        self.Initialization.setFixedWidth(150)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)
        self.grid.addWidget(self.createSecondExclusiveGroup(), 1, 0)
        self.grid.addWidget(self.createThirdExclusiveGroup(), 2, 0)
        # self.grid.addWidget(self.createForthExclusiveGroup(), 3, 0)
        self.grid.addWidget(self.createFifthExclusiveGroup(), 4, 0)

        self.setLayout(self.grid)
        self.do_Initialization()

    def __del__(self):
        print("Destructor called, Employee deleted.")

    def createFirstExclusiveGroup(self):
        self.groupBox_Parameters = QtWidgets.QGroupBox("Imaging setup parameters")

        self.grid1 = QtWidgets.QGridLayout()

        self.grid1.addWidget(self.wavelength_label, 0, 0)
        self.grid1.addWidget(self.wavelength, 1, 0)
        self.grid1.addWidget(self.NA_label, 0, 1)
        self.grid1.addWidget(self.NA, 1, 1)
        self.grid1.addWidget(self.ti0_label, 2, 0)
        self.grid1.addWidget(self.ti0, 3, 0)
        self.grid1.addWidget(self.ni0_label, 2, 1)
        self.grid1.addWidget(self.ni0, 3, 1)
        self.grid1.addWidget(self.ni_label, 4, 0)
        self.grid1.addWidget(self.ni, 5, 0)

        self.grid1.addWidget(self.tg0_label, 4, 1)
        self.grid1.addWidget(self.tg0, 5, 1)
        self.grid1.addWidget(self.tg_label, 0, 2)
        self.grid1.addWidget(self.tg, 1, 2)

        self.grid1.addWidget(self.ng0_label, 2, 2)
        self.grid1.addWidget(self.ng0, 3, 2)
        self.grid1.addWidget(self.ng_label, 4, 2)
        self.grid1.addWidget(self.ng, 5, 2)
        self.grid1.addWidget(self.ns_label, 0, 3)
        self.grid1.addWidget(self.ns, 1, 3)
        self.grid1.addWidget(self.pixel_size_physical_label, 2, 3)
        self.grid1.addWidget(self.pixel_size_physical, 3, 3)
        self.grid1.addWidget(self.pixel_size_label, 4, 3)
        self.grid1.addWidget(self.pixel_size, 5, 3)

        self.groupBox_Parameters.setLayout(self.grid1)
        return self.groupBox_Parameters

    def createSecondExclusiveGroup(self):
        self.groupBox_iPSF_model_1 = QtWidgets.QGroupBox("Modelling iPSF images (I) ")
        self.groupBox_iPSF_model_1.setCheckable(True)
        self.groupBox_iPSF_model_1.setChecked(True)
        self.groupBox_iPSF_model_1.toggled.connect(
            lambda: self.groupBox_iPSF_model_2.setChecked(False)
        )

        self.grid2 = QtWidgets.QGridLayout()
        self.grid2.addWidget(self.s_range_label, 0, 0)
        self.grid2.addWidget(self.s_range, 1, 0)
        self.grid2.addWidget(self.e_range_label, 2, 0)
        self.grid2.addWidget(self.e_range, 3, 0)
        self.grid2.addWidget(self.step_range_label, 4, 0)
        self.grid2.addWidget(self.step_range, 5, 0)
        self.grid2.addWidget(self.position_particle3D_label, 0, 1)
        self.grid2.addWidget(self.position_particle3D, 1, 1)
        self.grid2.addWidget(self.nx_label, 2, 1)
        self.grid2.addWidget(self.nx, 3, 1)
        self.grid2.addWidget(self.r_label, 4, 1)
        self.grid2.addWidget(self.r_, 5, 1)

        self.groupBox_iPSF_model_1.setLayout(self.grid2)
        return self.groupBox_iPSF_model_1

    def createThirdExclusiveGroup(self):
        self.groupBox_iPSF_model_2 = QtWidgets.QGroupBox("Modelling iPSF images (II)")
        self.groupBox_iPSF_model_2.setCheckable(True)
        self.groupBox_iPSF_model_2.setChecked(False)
        self.groupBox_iPSF_model_2.toggled.connect(
            lambda: self.groupBox_iPSF_model_1.setChecked(False)
        )

        self.grid3 = QtWidgets.QGridLayout()
        self.grid3.addWidget(self.s_range_label_particle, 0, 0)
        self.grid3.addWidget(self.s_range_particle, 1, 0)
        self.grid3.addWidget(self.e_range_label_particle, 2, 0)
        self.grid3.addWidget(self.e_range_particle, 3, 0)
        self.grid3.addWidget(self.step_range_label_particle, 4, 0)
        self.grid3.addWidget(self.step_range_particle, 5, 0)
        self.grid3.addWidget(self.nx_label_particle, 0, 1)
        self.grid3.addWidget(self.nx_particle, 1, 1)
        self.grid3.addWidget(self.r_label_particle, 2, 1)
        self.grid3.addWidget(self.r_particle, 3, 1)
        self.grid3.addWidget(self.z_focus_label_particle, 4, 1)
        self.grid3.addWidget(self.z_focus_particle, 5, 1)

        self.groupBox_iPSF_model_2.setLayout(self.grid3)
        return self.groupBox_iPSF_model_2

    # def createForthExclusiveGroup(self):
    #     self.groupBox_Parameters = QtWidgets.QGroupBox("Imaging setup parameters")
    #
    #     self.grid4 = QtWidgets.QGridLayout()
    #     self.grid4.addWidget(self.wavelength_label, 0, 0)
    #     self.groupBox_Parameters.setLayout(self.grid4)
    #     return self.groupBox_Parameters

    def createFifthExclusiveGroup(self):
        self.groupBox_PushButton = QtWidgets.QGroupBox("")

        self.grid5 = QtWidgets.QGridLayout()

        self.grid5.addWidget(self.generate_model, 0, 0)
        self.grid5.addWidget(self.Initialization, 0, 1)
        self.grid5.addWidget(self.display, 0, 2)

        self.groupBox_PushButton.setLayout(self.grid5)
        return self.groupBox_PushButton

    def get_values_Imaging(self):
        try:
            self.wavelength_var = float(self.wavelength.text())
            self.NA_var = float(self.NA.text())
            self.ti0_var = float(self.ti0.text())
            self.ni0_var = float(self.ni0.text())
            self.ni_var = float(self.ni.text())
            self.tg0_var = float(self.tg0.text())
            self.tg_var = float(self.tg.text())
            self.ng0_var = float(self.ng0.text())
            self.ng_var = float(self.ng.text())
            self.ns_var = float(self.ns.text())
            self.pixel_size_physical_var = float(self.pixel_size_physical.text())
            self.pixel_size_var = float(self.pixel_size.text())

            self.empty_value_ImagingBox_flag = True
        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all setup parameters!")
            self.msg_box3.exec_()

            self.empty_value_ImagingBox_flag = False

    def get_values_model(self):
        try:
            if self.groupBox_iPSF_model_1.isChecked():
                self.s_range_var = float(self.s_range.text())
                self.e_range_var = float(self.e_range.text())
                self.step_range_var = float(self.step_range.text())
                self.position_particle3D_var = eval(self.position_particle3D.text())
                self.nx_var = int(self.nx.text())
                self.r_var = int(self.r_.text())

                self.empty_value_ModelBox_flag = True
            elif self.groupBox_iPSF_model_2.isChecked():
                self.s_range_var = float(self.s_range_particle.text())
                self.e_range_var = float(self.e_range_particle.text())
                self.step_range_var = float(self.step_range_particle.text())
                self.nx_var = int(self.nx_particle.text())
                self.r_var = int(self.r_particle.text())
                self.z_focus_var = eval(self.z_focus_particle.text())
                self.empty_value_ModelBox_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all Modelling iPSF!")
            self.msg_box3.exec_()

            self.empty_value_ModelBox_flag = False

    def initial_values_Imaging(self):
        self.wavelength.setText("540e-9")
        self.NA.setText("1.4")
        self.ti0.setText("180e-6")
        self.ni0.setText("1.5")
        self.ni.setText("1.5")
        self.tg0.setText("170e-6")
        self.tg.setText("170e-6")
        self.ng0.setText("1.5")
        self.ng.setText("1.5")
        self.ns.setText("1.33")
        self.pixel_size_physical.setText("5.8e-6")
        self.pixel_size.setText("38e-9")

    def initial_values_model_I(self):
        self.s_range.setText("0")
        self.e_range.setText("10")
        self.step_range.setText("1e-1")
        self.position_particle3D.setText("[0, 0, 3e-6]")
        self.nx.setText("513")
        self.r_.setText("50")

    def initial_values_model_II(self):
        self.s_range_particle.setText("1.0")
        self.e_range_particle.setText("5.0")
        self.step_range_particle.setText("25e-3")
        self.position_particle3D.setText("[0, 0, 3e-6]")
        self.nx_particle.setText("513")
        self.r_particle.setText("50")
        self.z_focus_particle.setText("[3e-6]")

    def do_Initialization(self):
        self.initial_values_Imaging()
        self.initial_values_model_I()
        self.initial_values_model_II()

    def do_model(self):
        if self.generate_model.clicked:
            self.get_values_Imaging()
            self.get_values_model()
            if self.empty_value_ModelBox_flag and self.empty_value_ImagingBox_flag:
                p = ImagingSetupParameters

                # Wavelength of the light source in meters
                p.wavelength = self.wavelength_var
                # Numerical Aperture (NA) of the objective lens
                p.NA = self.NA_var
                # Thickness of the immersion oil, nominal value in meters
                p.ti0 = self.ti0_var
                # Refractive index of the immersion oil, nominal value
                p.ni0 = self.ni0_var
                # Refractive index of the immersion oil, experimental value
                p.ni = self.ni_var
                # Thickness of the coverglass, nominal value in meters
                p.tg0 = self.tg0_var
                # Thickness of the coverglass, experimental value in meters
                p.tg = self.tg_var
                # Refractive index of the coverglass, nominal value
                p.ng0 = self.ng0_var
                # Refractive index of the coverglass, experimental value
                p.ng = self.ng_var
                # Refractive index of the sample/medium
                p.ns = self.ns_var
                # Physical size of the camera pixel in meters
                p.pixel_size_physical = self.pixel_size_physical_var
                # Imaging pixel size in meters, related to the physical pixel
                # size through the magnification of the setup
                pixel_size = self.pixel_size_var
                # Magnification of the imaging system
                p.M = self.pixel_size_physical_var / self.pixel_size_var

                p.k0 = 2 * math.pi / self.wavelength_var  # Wavevector
                p.alpha = math.asin(
                    self.NA_var / self.ni_var
                )  # Largest angle collected by our Objective lens
                if self.groupBox_iPSF_model_1.isChecked():
                    z_focus_array = (
                        np.arange(self.s_range_var, self.e_range_var, self.step_range_var) * 1e-6
                    )
                    nz = np.size(z_focus_array)

                    scattered_field = ScatteredFieldDifferentialPhase(
                        p, self.position_particle3D_var, z_focus_array, nz, self.nx_var
                    )
                    (
                        scatteredFieldAmplitude_focalStack,
                        scatteredFieldPhase_focalStack,
                    ) = scattered_field.calculate(self.r_var)
                    iPSFs_focalStack = np.multiply(
                        scatteredFieldAmplitude_focalStack, np.cos(scatteredFieldPhase_focalStack)
                    )
                    self.iPSFs_focalStack = iPSFs_focalStack / np.max(
                        np.abs(iPSFs_focalStack[...])
                    )

                    list_titles = z_focus_array.tolist()
                    self.list_titles = [
                        "defocusing positions=" + str(z_ * 1e6) + " um" for z_ in list_titles
                    ]

                    self.iPSFs_focalStack_meirdonal = iPSFs_focalStack[:, 1 + self.r_var, :]
                    self.scatteredFieldAmplitude_focalStack_meirdonal = (
                        scatteredFieldAmplitude_focalStack[:, 1 + self.r_var, :]
                    )
                    self.scatteredFieldPhase_focalStack_meirdonal = (
                        scatteredFieldPhase_focalStack[:, 1 + self.r_var, :]
                    )

                    self.iPSFs_focalStack_meirdonal = (
                        self.iPSFs_focalStack_meirdonal / self.iPSFs_focalStack_meirdonal.max()
                    )
                    self.scatteredFieldAmplitude_focalStack_meirdonal = (
                        self.scatteredFieldAmplitude_focalStack_meirdonal
                        / self.scatteredFieldAmplitude_focalStack_meirdonal.max()
                    )
                    self.scatteredFieldPhase_focalStack_meirdonal = (
                        self.scatteredFieldPhase_focalStack_meirdonal
                        / self.scatteredFieldPhase_focalStack_meirdonal.max()
                    )

                elif self.groupBox_iPSF_model_2.isChecked():
                    nz = np.size(self.z_focus_var)
                    z_particle_array = (
                        np.arange(self.s_range_var, self.e_range_var, self.step_range_var) * 1e-6
                    )
                    self.scatteredFieldAmplitude_AxialStack = np.zeros(
                        (len(z_particle_array), 2 * self.r_var + 1, 2 * self.r_var + 1)
                    )
                    self.scatteredFieldPhase_AxialStack = np.zeros(
                        (len(z_particle_array), 2 * self.r_var + 1, 2 * self.r_var + 1)
                    )

                    for cnt, zp_ in enumerate(z_particle_array):
                        # The 3D position of the nanoparticle
                        Xp = [0, 0, zp_]
                        scattered_field = ScatteredFieldDifferentialPhase(
                            p, Xp, self.z_focus_var, nz, self.nx_var
                        )
                        scatteredFieldAmplitude, scatteredFieldPhase = scattered_field.calculate(
                            self.r_var
                        )
                        self.scatteredFieldAmplitude_AxialStack[
                            cnt, ...
                        ] = scatteredFieldAmplitude
                        self.scatteredFieldPhase_AxialStack[cnt, ...] = scatteredFieldPhase

                    iPSFs_AxialStack = np.multiply(
                        self.scatteredFieldAmplitude_AxialStack,
                        np.cos(self.scatteredFieldPhase_AxialStack),
                    )
                    self.iPSFs_AxialStack = iPSFs_AxialStack / np.max(
                        np.abs(iPSFs_AxialStack[...])
                    )

                    self.iPSFs_AxialStack_meirdonal = iPSFs_AxialStack[:, 1 + self.r_var, :]
                    self.scatteredFieldAmplitude_AxialStack_meirdonal = (
                        self.scatteredFieldAmplitude_AxialStack[:, 1 + self.r_var, :]
                    )
                    self.scatteredFieldPhase_AxialStack_meirdonal = (
                        self.scatteredFieldPhase_AxialStack[:, 1 + self.r_var, :]
                    )

                    self.iPSFs_AxialStack_meirdonal = (
                        self.iPSFs_AxialStack_meirdonal / self.iPSFs_AxialStack_meirdonal.max()
                    )
                    self.scatteredFieldAmplitude_AxialStack_meirdonal = (
                        self.scatteredFieldAmplitude_AxialStack_meirdonal
                        / self.scatteredFieldAmplitude_AxialStack_meirdonal.max()
                    )
                    self.scatteredFieldPhase_AxialStack_meirdonal = (
                        self.scatteredFieldPhase_AxialStack_meirdonal
                        / self.scatteredFieldPhase_AxialStack_meirdonal.max()
                    )

                    a = 1

    def do_display(self):
        if self.groupBox_iPSF_model_1.isChecked() and self.iPSFs_focalStack is not None:
            self.display_trigger.emit(["iPSF_Model_I", self.iPSFs_focalStack])
            self.pg = UpdatingPlots_Image(
                list_img=[
                    self.iPSFs_focalStack_meirdonal,
                    self.scatteredFieldAmplitude_focalStack_meirdonal,
                    self.scatteredFieldPhase_focalStack_meirdonal,
                ],
                list_titles=[
                    "Focal stack: iPSFs,",
                    "Scattered field amplitude,",
                    "Scattered field phase",
                ],
                x_axis_labels=[
                    "x-axis (x" + str(self.pixel_size_var * 1e9) + " nm)",
                    "x-axis (x" + str(self.pixel_size_var * 1e9) + " nm)",
                    "x-axis (x" + str(self.pixel_size_var * 1e9) + " nm)",
                ],
                y_axis_labels=[
                    "Focal stack: iPSFs,",
                    "Defocus axis (x" + str(1e2) + " nm)",
                    "Defocus axis (x" + str(1e2) + " nm)",
                ],
                title="iPSF_Model_I",
            )
        elif self.groupBox_iPSF_model_2.isChecked() and self.iPSFs_AxialStack is not None:
            self.display_trigger.emit(["iPSF_Model_II", self.iPSFs_AxialStack])
            self.pg = UpdatingPlots_Image(
                list_img=[
                    self.iPSFs_AxialStack_meirdonal,
                    self.scatteredFieldAmplitude_AxialStack_meirdonal,
                    self.scatteredFieldPhase_AxialStack_meirdonal,
                ],
                list_titles=[
                    "Axial stack: iPSFs,",
                    "Scattered field amplitude,",
                    "Scattered field phase",
                ],
                x_axis_labels=[
                    "x-axis (x" + str(self.pixel_size_var * 1e9) + " nm)",
                    "x-axis (x" + str(self.pixel_size_var * 1e9) + " nm)",
                    "x-axis (x" + str(self.pixel_size_var * 1e9) + " nm)",
                ],
                y_axis_labels=[
                    "Focal stack: iPSFs,",
                    "Defocus axis (x" + str(1e2) + " nm)",
                    "Defocus axis (x" + str(1e2) + " nm)",
                ],
                title="iPSF_Model_II",
            )

    @QtCore.Slot(int)
    def get_sliceNumber(self, frame_number):
        if self.groupBox_iPSF_model_1.isChecked():
            self.frame_num = frame_number
            x = list(range(0, self.iPSFs_focalStack_meirdonal.shape[0]))
            y = [frame_number for _ in range(len(x))]
            self.pg.update_plot(x, y)
            self.pg.show()

        elif self.groupBox_iPSF_model_2.isChecked():
            self.frame_num = frame_number
            x = list(range(0, self.iPSFs_AxialStack_meirdonal.shape[0]))
            y = [frame_number for _ in range(len(x))]
            self.pg.update_plot(x, y)
            self.pg.show()

        # print(self.list_titles[self.frame_num])



import os
import time
from functools import partial

import numpy as np
import pandas as pd
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Slot

from piscat.GUI.Projects.tab_bg_correction import BgCorrection_GUI
from piscat.GUI.Projects.tab_localization import Localization_GUI
from piscat.GUI.Projects.tab_plot_histogram import Histogram_GUI
from piscat.GUI.Visualization.fun_display_localization import Visulization_localization
from piscat.InputOutput import read_write_data
from piscat.InputOutput.read_write_data import load_dict_from_hdf5


class ProteinTabs(QtWidgets.QMainWindow):
    new_update_df_PSFs = QtCore.Signal(object)
    new_update_bg_video = QtCore.Signal(object)
    new_update_trajectories = QtCore.Signal(object)
    new_update_DRA_video = QtCore.Signal(object)

    def __init__(self, video_in, batch_size, object_update_progressBar, parent=None):
        super(ProteinTabs, self).__init__(parent)
        self.window = QtWidgets.QWidget()

        self.all_tabs = {}
        self.input_video = video_in
        self.batch_size = batch_size
        self.bgCorrectedVideo = None
        self.df_PSFs = None
        self.type_bgCorrection = None
        self.trajectories = None
        self.current_frame = 0
        self.object_update_progressBar = object_update_progressBar

        self.history = {
            "Type_bg_correction": None,
            "bg_setting": None,
            "PSFs_Localization_setting": None,
            "PSFs_Tracking_setting": None,
        }

        self.PSFs_Particels_num = {"#PSFs": None, "#Particles": None}

        self.initUI()

        self.setWindowTitle("Protein")
        self.window.setLayout(self.layout)
        self.setCentralWidget(self.window)
        self.window.show()

    def initUI(self):
        from piscat.GUI import Tracking_GUI

        self.resize(400, 200)
        self.setWindowTitle("PSF Tracking")
        self.setGeometry(1140, 100, 400, 200)

        # Tabs
        self.all_tabs = {
            "Bg_correction": BgCorrection_GUI(
                input_video=self.input_video,
                object_update_progressBar=self.object_update_progressBar,
            ),
            "PSFs_Localization": Localization_GUI(),
            "PSFs_Tracking": Tracking_GUI(self.batch_size),
            "Histogram": Histogram_GUI(),
        }

        self.all_tabs["Bg_correction"].output_Tab_bgCorrection.connect(
            self.Update_tab_bgCorrection
        )
        self.all_tabs["Bg_correction"].output_batchSize_Tab_bgCorrection.connect(
            self.Update_batchSize
        )
        self.all_tabs["Bg_correction"].output_setting_Tab_bgCorrection.connect(
            partial(self.history_update, key="bg_setting")
        )
        self.all_tabs["Bg_correction"].update_tab_index.connect(self.update_tab)

        self.all_tabs["PSFs_Localization"].preview_localization.connect(
            partial(self.Update_tab_localization, flag_preview=True)
        )
        self.all_tabs["PSFs_Localization"].update_localization.connect(
            partial(self.Update_tab_localization, flag_preview=False)
        )
        self.all_tabs["PSFs_Localization"].output_setting_Tab_Localization.connect(
            partial(self.history_update, key="PSFs_Localization_setting")
        )
        self.all_tabs["PSFs_Localization"].output_number_PSFs_tracking.connect(
            partial(self.number_PSFs, key="#PSFs")
        )
        self.all_tabs["PSFs_Localization"].update_tab_index.connect(self.update_tab)

        self.new_update_bg_video.connect(self.all_tabs["PSFs_Localization"].update_in_data)

        self.all_tabs["PSFs_Tracking"].update_tracking.connect(self.Update_tab_localization)
        self.all_tabs["PSFs_Tracking"].update_trajectories.connect(self.Update_tab_trajectories)
        self.all_tabs["PSFs_Tracking"].output_setting_Tab_tracking.connect(
            partial(self.history_update, key="PSFs_Tracking_setting")
        )
        self.all_tabs["PSFs_Tracking"].output_number_Particels_tracking.connect(
            partial(self.number_PSFs, key="#Particles")
        )
        self.new_update_df_PSFs.connect(self.all_tabs["PSFs_Tracking"].update_in_data)

        self.new_update_trajectories.connect(self.all_tabs["Histogram"].update_in_data)

        self.main_tabWidget = QtWidgets.QTabWidget()
        self.main_tabWidget.addTab(self.all_tabs["Bg_correction"], "Background correction")
        self.main_tabWidget.addTab(self.all_tabs["PSFs_Localization"], "PSFs localization")
        self.main_tabWidget.addTab(self.all_tabs["PSFs_Tracking"], "PSFs tracking")
        self.main_tabWidget.addTab(self.all_tabs["Histogram"], "Histogram")

        self.btn_save = QtWidgets.QPushButton("Save Data_History")
        self.btn_save.setAutoDefault(False)
        self.btn_save.clicked.connect(self.save_data)
        self.btn_save.setFixedWidth(120)
        self.btn_save.setFixedHeight(20)

        self.btn_load_csv = QtWidgets.QPushButton("Load CSV")
        self.btn_load_csv.setAutoDefault(False)
        self.btn_load_csv.clicked.connect(self.load_csv_data)
        self.btn_load_csv.setFixedWidth(100)
        self.btn_load_csv.setFixedHeight(20)

        self.btn_load_hdf5 = QtWidgets.QPushButton("Load hdf5")
        self.btn_load_hdf5.setAutoDefault(False)
        self.btn_load_hdf5.clicked.connect(self.load_hdf5_data)
        self.btn_load_hdf5.setFixedWidth(120)
        self.btn_load_hdf5.setFixedHeight(20)

        self.statusbar = self.statusBar()
        self.statusBar().addPermanentWidget(self.btn_save)
        self.statusBar().addPermanentWidget(self.btn_load_csv)
        self.statusBar().addPermanentWidget(self.btn_load_hdf5)

        self.main_tabWidget.setCurrentIndex(0)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.main_tabWidget)
        self.layout.setContentsMargins(0, 0, 0, 0)

    @Slot()
    def update_input(self, video_in):
        self.input_video = video_in

    def __del__(self):
        print("Destructor called, Employee deleted.")

    @Slot(int)
    def update_tab(self, idx_):
        self.main_tabWidget.setCurrentIndex(idx_)

    @Slot(int)
    def setProgress(self, val):
        self.progressBar.setValue(val)

    @Slot()
    def Update_batchSize(self, batchSize):
        self.batch_size = batchSize
        self.all_tabs["PSFs_Tracking"].update_batchSize(batchSize=batchSize)

    @Slot()
    def Update_tab_bgCorrection(self, data):
        if data[0] is not None:
            if len(data[0]) == 2:
                self.bgCorrectedVideo = data[0][0]
            else:
                self.bgCorrectedVideo = data[0]
            self.type_bgCorrection = data[1]
            self.history["Type_bg_correction"] = data[1]
            self.new_update_bg_video.emit(self.bgCorrectedVideo)

            if data[1] == "DRA":
                self.new_update_DRA_video.emit(
                    [self.bgCorrectedVideo, "DRA", None, self.batch_size]
                )

            self.visualization_ = Visulization_localization()
            self.visualization_.bg_correction_update(
                in_video=self.bgCorrectedVideo,
                label=self.type_bgCorrection,
                object=self.all_tabs["PSFs_Localization"],
            )
        else:
            self.msg_box2 = QtWidgets.QMessageBox()
            self.msg_box2.setWindowTitle("Warning!")
            self.msg_box2.setText("Input video does not find!")
            self.msg_box2.exec_()

    @QtCore.Slot()
    def Update_tab_localization(self, df_psfs, flag_preview=False):
        self.df_PSFs = df_psfs
        if self.bgCorrectedVideo is not None and self.df_PSFs is not None:
            if type(self.df_PSFs) is np.ndarray:
                self.visualization_.get_sliceNumber(
                    frame_num=self.all_tabs["PSFs_Localization"].frame_num
                )
                self.visualization_.update_localization_onMask(
                    video_in=self.bgCorrectedVideo,
                    title=self.type_bgCorrection,
                    df_PSFs=self.df_PSFs,
                )

            elif type(self.df_PSFs) is pd.core.frame.DataFrame:
                self.visualization_.get_sliceNumber(
                    frame_num=self.all_tabs["PSFs_Localization"].frame_num
                )
                self.visualization_.update_localization_onMask(
                    video_in=self.bgCorrectedVideo,
                    title=self.type_bgCorrection,
                    df_PSFs=self.df_PSFs,
                    flag_preview=flag_preview,
                )

                self.new_update_df_PSFs.emit([self.bgCorrectedVideo, self.df_PSFs])

        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Input Video is None!")
            self.msg_box1.exec_()

        elif self.df_PSFs is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Zero particle detected!")
            self.msg_box1.exec_()

    @QtCore.Slot()
    def Update_tab_trajectories(self, trajectories):
        self.trajectories = trajectories
        self.new_update_trajectories.emit(
            [trajectories, self.input_video.shape[0], self.batch_size]
        )

    @QtCore.Slot()
    def history_update(self, history, key):
        self.history[key] = history

    @QtCore.Slot()
    def number_PSFs(self, data_in, key):
        self.PSFs_Particels_num[key] = data_in

    def save_data(self):
        self.file_path = False
        self.file_path = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select Folder", os.path.expanduser("~"), QtWidgets.QFileDialog.ShowDirsOnly
        )
        timestr = time.strftime("%Y%m%d-%H%M%S")
        name_mkdir = timestr
        try:
            dr_mk = os.path.join(self.file_path, name_mkdir)
            os.mkdir(dr_mk)
            print("Directory ", name_mkdir, " Created ")
        except FileExistsError:
            dr_mk = os.path.join(self.file_path, name_mkdir)
            print("Directory ", name_mkdir, " already exists")

        self.file_path = dr_mk

        if self.file_path:
            if self.df_PSFs is not None:
                read_write_data.save_df2csv(self.df_PSFs, path=self.file_path, name="df_PSFs")

            fout = os.path.join(self.file_path, "history.txt")
            fo = open(fout, "w")
            for k, v in self.history.items():
                fo.write(str(k) + " >>> " + str(v) + "\n\n")

            for k, v in self.PSFs_Particels_num.items():
                fo.write(str(k) + " >>> " + str(v) + "\n\n")
            fo.close()

            if self.trajectories is not None:
                read_write_data.save_mat(
                    data=self.trajectories, path=self.file_path, name="all_trajectories"
                )
                read_write_data.save_list_to_hdf5(
                    list_data=self.trajectories, path=self.file_path, name="histData"
                )

    def load_hdf5_data(self):
        if self.bgCorrectedVideo is not None:
            try:
                fileName = QtWidgets.QFileDialog().getOpenFileName()
                filePath = str(fileName[0])
                data_dic = load_dict_from_hdf5(filePath)
                self.Update_tab_trajectories(data_dic)
            except:
                print("HDF5 can not load!")
        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Update background tab!")
            self.msg_box1.exec_()

    def load_csv_data(self):
        if self.bgCorrectedVideo is not None:
            try:
                fileName = QtWidgets.QFileDialog().getOpenFileName()
                filePath = str(fileName[0])
                df_PSFs = pd.read_csv(filePath)
                self.Update_tab_localization(df_PSFs)
            except:
                print("CSV can not load!")
        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Update background tab!")
            self.msg_box1.exec_()
